import numpy as np
import pytest
from numpy.testing import assert_array_equal

from cchdo.hydro import accessors  # noqa: F401
from cchdo.hydro.exchange.exceptions import ExchangeDataFlagPairError


def test_fq_merge(nc_placeholder):
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "1",
            "OXYGEN [UMOL/KG]": "234.1",
            "OXYGEN [UMOL/KG]_FLAG_W": "2",
        }
    ]
    merged = nc_placeholder.cchdo.merge_fq(fq)

    assert merged.oxygen.data == [[234.1]]
    assert merged.oxygen_qc.data == [[2]]
    assert merged.oxygen.attrs["C_format"] == "%.1f"
    assert merged.oxygen.attrs["C_format_source"] == "input_file"


def test_fq_merge_with_error(nc_placeholder):
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "1",
            "DELC14 [/MILLE]": "-3.12",
            "DELC14 [/MILLE]_FLAG_W": "2",
            "C14ERR [/MILLE]": "0.12",
        }
    ]
    merged = nc_placeholder.cchdo.merge_fq(fq)

    assert merged.del_carbon_14_dic.data == [[-3.12]]
    assert merged.del_carbon_14_dic_qc.data == [[2]]
    assert merged.del_carbon_14_dic_error.data == [[0.12]]
    assert merged.del_carbon_14_dic_error.attrs["C_format"] == "%.2f"
    assert merged.del_carbon_14_dic_error.attrs["C_format_source"] == "input_file"


def test_fq_merge_flag_error(nc_placeholders):
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "OXYGEN [UMOL/KG]": "200.1",
        }
    ]
    with pytest.raises(ExchangeDataFlagPairError):
        nc_placeholders.cchdo.merge_fq(fq)


def test_fq_merge_flag_only(nc_placeholders):
    # Ensure that a valid flag only update succeeds
    # 9 -> 1 for missing data
    # 1 -> 9 for missing data
    # 1 -> 5 for missing data
    # 2 -> 3 in the case of extant data

    # 9 -> 1
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "OXYGEN [UMOL/KG]_FLAG_W": "1",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    assert_array_equal(merged.oxygen_qc.data, [[1, 1, 5, np.nan, 1, 2]])

    # 1 -> 9
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "35",
            "OXYGEN [UMOL/KG]_FLAG_W": None,
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    assert_array_equal(merged.oxygen_qc.data, [[np.nan, np.nan, 5, np.nan, 1, 2]])

    # 1 -> 5
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "35",
            "OXYGEN [UMOL/KG]_FLAG_W": "5",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    assert_array_equal(merged.oxygen_qc.data, [[np.nan, 5, 5, np.nan, 1, 2]])

    # 2 -> 3
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "30",
            "OXYGEN [UMOL/KG]_FLAG_W": "3",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    assert_array_equal(merged.oxygen_qc.data, [[np.nan, 1, 5, np.nan, 1, 3]])


def test_fq_merge_cdom(nc_placeholders):
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "CDOM300 [/METER]": "100.2",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    arr = np.full((6, 2), fill_value=np.nan)
    arr[0][0] = 100.2
    assert_array_equal(merged.cdom, [arr])

    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "CDOM325 [/METER]": "100.2",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)
    arr = np.full((6, 2), fill_value=np.nan)
    arr[0][1] = 100.2
    assert_array_equal(merged.cdom, [arr])


def test_fq_merge_with_alt_params(nc_placeholders):
    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "SILCAT [UMOL/KG]": "100.1",
            "SILCAT [UMOL/KG]_FLAG_W": "2",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)

    assert_array_equal(
        merged.silicate.data, [[100.1, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_qc.data, [[2, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_alt_1.data, [[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_alt_1_qc.data,
        [[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]],
    )

    fq = [
        {
            "EXPOCODE": "TEST",
            "STNNBR": "1",
            "CASTNO": 1,
            "SAMPNO": "36",
            "SILCAT_ALT_1 [UMOL/KG]": "100.1",
            "SILCAT_ALT_1 [UMOL/KG]_FLAG_W": "2",
            "SILUNC_ALT_1 [UMOL/KG]": "1.0",
        }
    ]
    merged = nc_placeholders.cchdo.merge_fq(fq)

    assert_array_equal(
        merged.silicate_alt_1.data, [[100.1, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_alt_1_qc.data, [[2, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_alt_1_error.data,
        [[1.0, np.nan, np.nan, np.nan, np.nan, np.nan]],
    )
    assert_array_equal(
        merged.silicate.data, [[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_qc.data, [[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )
    assert_array_equal(
        merged.silicate_error.data, [[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]]
    )



import cchdo.hydro.legacy.woce as legacy_woce


def test_woce_flag_defs():
    """This test was made to ensure that a refactor produced the same result"""

    BOTTLE_FLAG_DESCRIPTION = "::1 = Bottle information unavailable.:2 = No problems noted.:3 = Leaking.:4 = Did not trip correctly.:5 = Not reported.:6 = Significant discrepancy in measured values between Gerard and Niskin bottles.:7 = Unknown problem.:8 = Pair did not trip correctly. Note that the Niskin bottle can trip at an unplanned depth while the Gerard trips correctly and vice versa.:9 = Samples not drawn from this bottle.:\n"
    CTD_FLAG_DESCRIPTION = "::1 = Not calibrated:2 = Acceptable measurement:3 = Questionable measurement:4 = Bad measurement:5 = Not reported:6 = Interpolated over >2 dbar interval:7 = Despiked:8 = Not assigned for CTD data:9 = Not sampled:\n"
    WATER_SAMPLE_FLAG_DESCRIPTION = "::1 = Sample for this measurement was drawn from water bottle but analysis not received.:2 = Acceptable measurement.:3 = Questionable measurement.:4 = Bad measurement.:5 = Not reported.:6 = Mean of replicate measurements.:7 = Manual chromatographic peak measurement.:8 = Irregular digital chromatographic peak integration.:9 = Sample not drawn for this measurement from this bottle.:\n"

    assert legacy_woce.BOTTLE_FLAG_DESCRIPTION == BOTTLE_FLAG_DESCRIPTION
    assert legacy_woce.CTD_FLAG_DESCRIPTION == CTD_FLAG_DESCRIPTION
    assert legacy_woce.WATER_SAMPLE_FLAG_DESCRIPTION == WATER_SAMPLE_FLAG_DESCRIPTION



import json
import warnings
from importlib.resources import read_text
from io import BytesIO
from zipfile import ZipFile

import pytest
import xarray as xr
from xarray.testing import assert_identical

from cchdo.hydro.accessors import CCHDOAccessor
from cchdo.hydro.exchange import FileType, read_csv, read_exchange

exp_stn_cast = json.loads(read_text("cchdo.hydro.tests.data", "stns_test_data.json"))


@pytest.mark.parametrize("expocode", ["318M20130321", "320620140320", "33KI159/1"])
@pytest.mark.parametrize("station", exp_stn_cast["stations"])
@pytest.mark.parametrize("cast", [0, 1, 10, 100, 999, 9999])
@pytest.mark.parametrize("profile_type", [FileType.BOTTLE, FileType.CTD])
@pytest.mark.parametrize("profile_count", [1, 2])
@pytest.mark.parametrize("ftype", ["cf", "exchange", "coards", "woce"])
def test_gen_fname_machinery(
    expocode, station, cast, profile_type, profile_count, ftype
):
    CCHDOAccessor._gen_fname(
        expocode, station, cast, profile_type, profile_count, ftype
    )


def test_coards_no_comments():
    # note that the differet bottle and sampno are intentional
    test_data = b"""BOTTLE,test
# some comment
EXPOCODE,STNNBR,CASTNO,SAMPNO,BTLNBR,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS
,,,,,,,,,DBAR
TEST          ,1       ,  1,1          ,2          ,20200101,0000,        0,        0,        0
END_DATA
"""
    ds = read_exchange(test_data)
    ds.attrs["comments"] = ""
    ds.cchdo.to_coards()


def test_exchange_bottle_round_trip():
    # note that the differet bottle and sampno are intentional
    test_data = b"""BOTTLE,test
# some comment
EXPOCODE,STNNBR,CASTNO,SAMPNO,BTLNBR,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS
,,,,,,,,,DBAR
TEST          ,1       ,  1,1          ,2          ,20200101,0000,        0,        0,        0
END_DATA
"""
    ds = read_exchange(test_data)
    # the magic slice removes the stamp and the newline with #
    rt = read_exchange(ds.cchdo.to_exchange()[25:])
    assert_identical(ds, rt)


def test_exchange_bottle_round_trip_with_alt():
    # note that the differet bottle and sampno are intentional
    test_data = b"""BOTTLE,test
# some comment
EXPOCODE,STNNBR,CASTNO,SAMPNO,BTLNBR,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS,CTDSAL,CTDSAL_ALT_1,CTDSAL_ALT_1_FLAG_W,SILCAT,SILCAT_ALT_1,SILUNC_ALT_1
,,,,,,,,,DBAR,PSS-78,PSS-78,,UMOL/KG,UMOL/KG,UMOL/KG
TEST          ,1       ,  1,1          ,2          ,20200101,0000,        0,        0,        0,35,36,2,10,10,3
END_DATA
"""
    ds = read_exchange(test_data)
    # the magic slice removes the stamp and the newline with #
    rt = read_exchange(ds.cchdo.to_exchange()[25:])
    assert_identical(ds, rt)


def test_exchange_bottle_round_trip_cdom():
    # note that the differet bottle and sampno are intentional
    # note that we do want the print precision of the two CDOMs to differ
    test_data = b"""BOTTLE,test
# some comment
EXPOCODE,STNNBR,CASTNO,SAMPNO,BTLNBR,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS,CDOM-300,CDOM-300_FLAG_W,CDOM-325,CDOM-325_FLAG_W
,,,,,,,,,DBAR,/METER,,/METER,
TEST          ,1       ,  1,1          ,2          ,20200101,0000,        0,        0,        0,      0.052,2,     0.0091,2
END_DATA
"""
    ds = read_exchange(test_data)
    assert "N_PROF" not in ds.variables
    assert "N_LEVELS" not in ds.variables
    assert "cdom300" not in ds
    assert "cdom325" not in ds
    assert "cdom" in ds
    # the magic slice removes the stamp and the newline with #
    rt = read_exchange(ds.cchdo.to_exchange()[25:])
    assert_identical(ds, rt)


def test_exchange_ctd_round_trip():
    test_data = b"""CTD,test
# some comment
NUMBER_HEADERS = 8
EXPOCODE = TEST
STNNBR = 1
CASTNO = 1
DATE = 20200101
TIME = 0000
LATITUDE = 0
LONGITUDE = 0
CTDPRS
DBAR
0
END_DATA
"""
    ds = read_exchange(test_data)
    # the magic slice removes the stamp and the newline with #
    rt = read_exchange(ds.cchdo.to_exchange()[22:])
    assert_identical(ds, rt)


def test_nc_serialize_all_ctd(tmp_path):
    """A crash was discovered when the ctd elapsed time param was present, and was seralized to disk then read back in"""
    test_data = b"""CTD,test
# some comment
NUMBER_HEADERS = 8
EXPOCODE = TEST
STNNBR = 1
CASTNO = 1
DATE = 20200101
TIME = 0000
LATITUDE = 0
LONGITUDE = 0
CTDPRS,CTDTMP,CTDTMP_FLAG_W
DBAR,ITS-90,
0,0,2
END_DATA
"""
    ds = read_exchange(test_data)
    nc = tmp_path / "test.nc"
    ds.to_netcdf(nc)
    ds = xr.load_dataset(nc)
    # the magic slice removes the stamp and the newline with #
    ds.cchdo.to_exchange()
    ds.cchdo.to_coards()
    ds.cchdo.to_woce()
    ds.cchdo.to_sum()


def test_nc_serialize_all_ctdetime(tmp_path):
    """A crash was discovered when the ctd elapsed time param was present, and was seralized to disk then read back in"""
    test_data = b"""CTD,test
# some comment
NUMBER_HEADERS = 8
EXPOCODE = TEST
STNNBR = 1
CASTNO = 1
DATE = 20200101
TIME = 0000
LATITUDE = 0
LONGITUDE = 0
CTDPRS,CTDETIME
DBAR,SECONDS
0,-999
END_DATA
"""
    ds = read_exchange(test_data)
    nc = tmp_path / "test.nc"
    ds.to_netcdf(nc)
    ds = xr.load_dataset(nc)
    # the magic slice removes the stamp and the newline with #
    ds.cchdo.to_exchange()
    ds.cchdo.to_coards()
    ds.cchdo.to_woce()
    ds.cchdo.to_sum()


def test_woce_ctd_no_flags(tmp_path):
    """make sure data is written to the woce files when there are no qc flags"""
    test_data = b"""EXPOCODE,SECT_ID,STNNBR,CASTNO,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR],CTDTMP [DEG C],CTDSAL [PSS-78],CTDOXY [UMOL/KG],CTDFLUOR [MG/M^3],CTDBEAMCP [/METER]
64PE20110724,AR07E,9,1,20110729,1919,59.57017,-38.77183,3006.0,1.2096,34.8913,304.0,0.0,0.161
"""
    ds = read_csv(test_data, ftype="C")
    zip_data = BytesIO(ds.cchdo.to_woce())
    with ZipFile(zip_data) as zf:
        fname = zf.namelist()[0]
        with zf.open(fname) as ctd:
            # this asserts that the above data block exists somewhere in the resulting file
            # the previous bug would just have a blank data block section
            assert b"3006.0  1.2096 34.8913   304.0" in ctd.read()


def test_coards_ctdnobs_with_missing():
    """Test a condition where exception might be thrown when a float array with nans is cast to int

    This is very platform dependent and depends on undefined behavior in C. Lucky for us we can see if
    numpy is issuing a runtime warning (they are considering makeing this throw in the future anyway)
    """
    test_data = b"""EXPOCODE,SECT_ID,STNNBR,CASTNO,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR],CTDTMP [DEG C],CTDSAL [PSS-78],CTDNOBS
64PE20110724,AR07E,9,1,20110729,1919,59.57017,-38.77183,3006.0,1.2096,34.8913,-999
"""
    ds = read_csv(test_data, ftype="C")
    with warnings.catch_warnings(record=True) as w:
        ds.cchdo.to_coards()
        for warning in w:
            if isinstance(warning.message, str):
                assert "invalid value encountered in cast" not in warning.message


def test_coards_bottle_flag_9():
    """Test a condition where a warning might be thrown when a bottle has flag 9"""
    test_data = b"""EXPOCODE,SECT_ID,STNNBR,CASTNO,SAMPNO,BTLNBR,BTLNBR_FLAG_W,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR]
64PE20110724,AR07E,9,1,1,1,9,20110729,1919,59.57017,-38.77183,3006.0,0
"""
    ds = read_csv(test_data, ftype="B")
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter(action="ignore", category=FutureWarning)
        ds.cchdo.to_coards()
        # simply assert that no warnings were issued durring the test
        # numpy would issue a RuntimeWarning if an unsafe cast occured
        assert len(w) == 0


def test_nc_serialize_all_ctdetime_multiple_profs(tmp_path):
    """Test file with CTDETIME and differing profile lengths

    The exchange writer would crash with a zip strict error, this is solved by setting decode_timedelta to False when the file is read in
    """
    test_data = b"""EXPOCODE,STNNBR,CASTNO,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR],CTDETIME[SECONDS]
TEST,1,1,20200101,0000,0,0,0,30
TEST,1,1,20200101,0000,0,0,2,1
TEST,2,1,20200101,0000,0,0,0,1
"""
    ds = read_csv(test_data, ftype="C")
    nc = tmp_path / "test.nc"
    ds.to_netcdf(nc)
    ds = xr.load_dataset(nc, decode_timedelta=None)
    ds.cchdo.to_exchange()
    ds.cchdo.to_coards()
    ds.cchdo.to_woce()
    ds.cchdo.to_sum()

    ds = xr.load_dataset(nc, decode_timedelta=False)
    ds.cchdo.to_exchange()
    ds.cchdo.to_coards()
    ds.cchdo.to_woce()
    ds.cchdo.to_sum()


def test_ctdetime_preserve_precision(tmp_path):
    """Test file with CTDETIME and differing profile lengths

    The exchange writer would crash with a zip strict error, this is solved by setting decode_timedelta to False when the file is read in
    """
    test_data = b"""EXPOCODE,STNNBR,CASTNO,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR],CTDETIME[SECONDS]
TEST,2,1,20200101,0000,0,0,0,3.1415
"""
    ds = read_csv(test_data, ftype="C")
    nc = tmp_path / "test.nc"
    assert ds.ctd_elapsed_time.attrs["C_format"] == "%.4f"
    ds.to_netcdf(nc)
    ds = xr.load_dataset(nc, decode_timedelta=False)
    # the magic slice removes the stamp and the newline with #
    ex = ds.cchdo.to_exchange()
    assert b"3.1415" in ex



from collections import defaultdict

import numpy as np
import xarray as xr

from cchdo.hydro.exchange.exceptions import ExchangeDataFlagPairError
from cchdo.hydro.flags import (
    ExchangeBottleFlag,
    ExchangeCTDFlag,
    ExchangeSampleFlag,
)
from cchdo.hydro.sorting import sort_ds


def check_sorted(dataset: xr.Dataset) -> bool:
    """Check that the dataset is sorted by the rules in :func:`sort_ds`"""
    sorted_ds = sort_ds(dataset.copy(deep=True))

    return all(
        [
            np.allclose(sorted_ds.pressure, dataset.pressure, equal_nan=True),
            np.all(
                (sorted_ds.time == dataset.time)
                | (np.isnat(sorted_ds.time) == np.isnat(dataset.time))
            ),
            np.allclose(sorted_ds.latitude, dataset.latitude, equal_nan=True),
            np.allclose(sorted_ds.longitude, dataset.longitude, equal_nan=True),
        ]
    )


def check_ancillary_variables(ds: xr.Dataset):
    """Check that everything in an ancillary_variables attribute appears as a variable
    Check that every variable that is known ancillary appears in at least one ancillary_variable attribute
    """
    looks_ancillary_suffixes = ("_qc", "_error")

    ancillary_variables_attrs = defaultdict(list)
    looks_ancillary = set()

    for name, variable in ds.variables.items():
        if not isinstance(name, str):
            raise ValueError(f"variable names must be strings not {name}")

        if any(name.endswith(suffix) for suffix in looks_ancillary_suffixes):
            looks_ancillary.add(name)

        if variable.attrs.get("ancillary_variables") is None:
            continue

        for ancillary in variable.attrs["ancillary_variables"].split():
            ancillary_variables_attrs[ancillary].append(name)

    if errors := ancillary_variables_attrs.keys() - ds.variables.keys():
        raise ValueError(errors)

    if errors := looks_ancillary - ancillary_variables_attrs.keys():
        raise ValueError(errors)


def check_flags(dataset: xr.Dataset, raises=True):
    """Check WOCE flag values agaisnt their param and ensure that the param either has a value or is "nan" depedning on the flag definition.

    Return a boolean array of invalid locations?
    """
    woce_flags = {
        "WOCESAMPLE": ExchangeBottleFlag,
        "WOCECTD": ExchangeCTDFlag,
        "WOCEBOTTLE": ExchangeSampleFlag,
    }
    flag_has_value = {
        "WOCESAMPLE": {flag.value: flag.has_value for flag in ExchangeBottleFlag},
        "WOCECTD": {flag.value: flag.has_value for flag in ExchangeCTDFlag},
        "WOCEBOTTLE": {flag.value: flag.has_value for flag in ExchangeSampleFlag},
    }
    # In some cases, a coordinate variable might have flags, so we are not using filter_by_attrs
    # get all the flag vars (that also have conventions)
    flag_vars = []
    for var_name in dataset.variables:
        # do not replace the above with .items() it will give you xr.Variable objects that you don't want to use
        # the following gets a real xr.DataArray
        data = dataset[var_name]
        if not {"standard_name", "conventions"} <= data.attrs.keys():
            continue
        if not any(flag in data.attrs["conventions"] for flag in woce_flags):
            continue
        if "status_flag" in data.attrs["standard_name"]:
            flag_vars.append(var_name)

    # match flags with their data vars
    # it is legal in CF for one set of flags to apply to multiple vars
    flag_errors = {}
    for flag_var in flag_vars:
        # get the flag and check attrs for defs
        flag_da = dataset[flag_var]
        conventions = None
        for flag in woce_flags:
            if flag_da.attrs.get("conventions", "").startswith(flag):
                conventions = flag
                break

        # we don't know these flags, skip the check
        if not conventions:
            continue

        allowed_values = np.array(list(flag_has_value[conventions]))
        illegal_flags = ~flag_da.fillna(9).isin(allowed_values)
        if np.any(illegal_flags):
            illegal_flags.attrs["comments"] = (
                f"This is a boolean array in the same shape as '{flag_da.name}' which is truthy where invalid values exist"
            )
            flag_errors[f"{flag_da.name}_value_errors"] = illegal_flags
            continue

        for var_name in dataset.variables:
            data = dataset[var_name]
            if "ancillary_variables" not in data.attrs:
                continue
            if flag_var not in data.attrs["ancillary_variables"].split(" "):
                continue

            # check data against flags
            has_fill_f = [
                flag
                for flag, value in flag_has_value[conventions].items()
                if value is False
            ]

            has_fill = flag_da.isin(has_fill_f) | np.isnan(flag_da)

            # TODO deal with strs

            if np.issubdtype(data.values.dtype, np.number):
                fill_value_mismatch: xr.DataArray = ~(np.isfinite(data) ^ has_fill)  # type: ignore[assignment]
                if np.any(fill_value_mismatch):
                    fill_value_mismatch.attrs["comments"] = (
                        f"This is a boolean array in the same shape as '{data.name}' which is truthy where invalid values exist"
                    )
                    flag_errors[f"{data.name}_value_errors"] = fill_value_mismatch

    flag_errors_ds = xr.Dataset(flag_errors)
    if raises and any(flag_errors_ds):
        raise ExchangeDataFlagPairError(flag_errors_ds)

    return flag_errors_ds



import numpy as np
import xarray as xr


def sort_ds(dataset: xr.Dataset) -> xr.Dataset:
    """Sorts the data values in the dataset

    Ensures that profiles are in the following order:

    * Earlier before later (time will increase)
    * Southerly before northerly (latitude will increase)
    * Westerly before easterly (longitude will increase)

    The two xy sorts are esentially tie breakers for when we are missing "time"

    Inside profiles:

    * Shallower before Deeper (pressure will increase)

    """
    # first make sure everything is sorted by pressure
    # this is being done "manually" here becuase xarray only supports 1D sorting
    pressure = dataset.pressure
    sorted_indicies = np.argsort(pressure.values, axis=1)

    for var in dataset.variables:
        # this check ensures that the variable being sorted
        # shares the first two dims as pressure, but allows for more dims past that
        if dataset[var].dims[slice(0, len(pressure.dims))] == pressure.dims:
            dataset[var][:] = np.take_along_axis(
                dataset[var].values, sorted_indicies, axis=1
            )

    # now we can just use the xarray sorting, which only supports 1D
    return dataset.sortby(["time", "latitude", "longitude"])



""" """

from collections.abc import Mapping

from xarray import Dataset


def is_not_none(obj):
    return obj is not None


def rename_with_bookkeeping(
    xarray_obj: Dataset,
    name_dict: Mapping | None = None,
    attrs: list[str] | None = None,
) -> Dataset:
    """Find and update all instances of a given variable to a new name.

    Parameters can be referenced in the attributes of separate parameter (e.g. ``ancillary_variables``) and need to be updated appropriately when renaming variables.

    :param xarray.Dataset xarray_obj: A Dataset containing variables, flags, etc.
    :param typing.Mapping name_dict: Mapping of old variable names to new.
    :param typing.List[str] attrs: Names of variable attributes to search through.
    :rtype: xarray.Dataset
    """
    # lets just noop this case
    if name_dict is None:
        return xarray_obj

    # easy part
    renamed = xarray_obj.rename(name_dict)

    # no bookkeeping to do
    if attrs is None:
        return renamed

    # find and search through all variables which have attr
    for attr in attrs:
        for var, ds in renamed.filter_by_attrs(**{attr: is_not_none}).items():
            attr_values = ds.attrs[attr].split(" ")
            for key, value in name_dict.items():
                for idx, attr_value in enumerate(attr_values):
                    if key == attr_value:
                        attr_values[idx] = attr_value.replace(key, value)
            renamed[var].attrs[attr] = " ".join(attr_values)

    return renamed


def to_argo_variable_names(xarray_obj: Dataset) -> Dataset:
    return xarray_obj



from datetime import datetime, timedelta
from warnings import warn

import numpy as np
import numpy.typing as npt
import xarray as xr

from cchdo.hydro.exchange.exceptions import ExchangeError
from cchdo.hydro.types import WHPNameAttr
from cchdo.params import WHPName, WHPNames

DATE = WHPNames["DATE"]
TIME = WHPNames["TIME"]


def _combine_dt_ndarray(
    date_arr: npt.NDArray[np.str_],
    time_arr: npt.NDArray[np.str_] | None = None,
    time_pad=False,
) -> np.ndarray:
    # TODO: When min pyver is 3.10, maybe consider pattern matching here
    def _parse_date(date_val: str) -> np.datetime64:
        if date_val == "":
            return np.datetime64("nat")
        return np.datetime64(datetime.strptime(date_val, "%Y%m%d"))

    def _parse_datetime(date_val: str) -> np.datetime64:
        if date_val == "T":
            return np.datetime64("nat")
        if date_val.endswith("2400"):
            date, _ = date_val.split("T")
            return np.datetime64(datetime.strptime(date, "%Y%m%d") + timedelta(days=1))
        return np.datetime64(datetime.strptime(date_val, "%Y%m%dT%H%M"))

    # vectorize here doesn't speed things, it just nice for the interface
    parse_date = np.vectorize(_parse_date, ["datetime64"])
    parse_datetime = np.vectorize(_parse_datetime, ["datetime64"])

    if time_arr is None:
        return parse_date(date_arr).astype("datetime64[D]")

    if np.all(time_arr == "0"):
        return parse_date(date_arr).astype("datetime64[D]")

    time_arr = time_arr.astype("U4")

    if time_pad:
        if np.any(np.char.str_len(time_arr[time_arr != ""]) < 4):
            warn("Time values are being padded with zeros", stacklevel=2)
        if not np.all(time_arr == ""):
            time_arr[time_arr != ""] = np.char.zfill(time_arr[time_arr != ""], 4)

    arr = np.char.add(np.char.add(date_arr, "T"), time_arr)
    return parse_datetime(arr).astype("datetime64[m]")


def combine_dt(
    dataset: xr.Dataset,
    is_coord: bool = True,
    date_name: WHPName = DATE,
    time_name: WHPName = TIME,
    time_pad=False,
) -> xr.Dataset:
    """Combine the exchange style string variables of date and optinally time into a single
    variable containing real datetime objects

    This will remove the time variable if present, and replace then rename the date variable.
    Date is replaced/renamed to maintain variable order in the xr.DataSet
    """

    # date and time want specific attrs whos values have been
    # selected by significant debate
    date = dataset[date_name.full_nc_name]
    time: xr.DataArray | None = dataset.get(
        time_name.full_nc_name
    )  # not be present, this is allowed

    whp_name: WHPNameAttr = [date_name.full_whp_name, time_name.full_whp_name]
    try:
        if time is None:
            dt_arr = _combine_dt_ndarray(date.values)
        else:
            dt_arr = _combine_dt_ndarray(date.values, time.values, time_pad=time_pad)
    except ValueError as err:
        raise ExchangeError(
            f"Could not parse date/time cols {date_name.whp_name} {time_name.whp_name}"
        ) from err

    precision = 1 / 24 / 60  # minute as day fraction
    if dt_arr.dtype.name == "datetime64[D]":
        precision = 1
        whp_name = date_name.whp_name

    time_var = xr.DataArray(
        dt_arr.astype("datetime64[ns]"),
        dims=date.dims,
        attrs={
            "standard_name": "time",
            "whp_name": whp_name,
            "resolution": precision,
        },
    )
    if is_coord is True:
        time_var.attrs["axis"] = "T"

    # if the thing being combined is a coordinate, it may not contain vill values
    time_var.encoding["_FillValue"] = None if is_coord else np.nan
    time_var.encoding["units"] = "days since 1950-01-01T00:00Z"
    time_var.encoding["calendar"] = "gregorian"
    time_var.encoding["dtype"] = "double"

    try:
        del dataset[time_name.nc_name]
    except KeyError:
        pass

    # this is being done in a funny way to retain the variable ordering
    # we will always keep the "time" variable name
    dataset[date_name.nc_name] = time_var
    return dataset.rename({date_name.nc_name: time_name.nc_name})



"""Legacy COARDS netcdf make from libcchdo ported to take a CCHDO CF/netCDF xarray.Dataset object as input.

The goal is, as much as possible, to use the old code with minimal changes such that the following outputs are identical:

* Exchange -> CF/netCDF -> COARDS netCDF (this library)
* Exchange -> COARDS netCDF (using libcchdo)

The entrypoint function is :func:`to_coards`
"""

import datetime
from csv import DictReader

# TODO: switch to files().joinpath().open when python 3.8 is dropped
# 2023-04-16
from importlib.resources import open_text
from io import BytesIO
from logging import getLogger
from tempfile import TemporaryDirectory
from typing import Literal
from zipfile import ZIP_DEFLATED, ZipFile

import numpy as np
import xarray as xr
from cftime import date2num

from cchdo.hydro import accessors as acc
from cchdo.hydro.legacy import woce

log = getLogger(__name__)
"""logger object for message logging"""

# Load the WHP to nc name mapping
# this was dumped from old libcchdo
with open_text("cchdo.hydro.legacy.coards", "name_netcdf.csv") as params:
    PARAMS = {}
    """mapping of whp names to nc names

    This is loaded at module import time from a dump from the old internal params sqlite database
    """
    for param in DictReader(params):
        PARAMS[f"{param['name']} [{param['mnemonic']}]"] = param

# collection exts
CTD_ZIP_FILE_EXTENSION = "nc_ctd.zip"
"""Filename extention for a zipped collection ctd coards netcdf files"""
BOTTLE_ZIP_FILE_EXTENSION = "nc_hyd.zip"
"""Filename extention for a zipped collection bottle coards netcdf files"""

# These consts are taken directly from libcchdo
FILL_VALUE = -999.0
"""Const from old libcchdo, -999.0"""

QC_SUFFIX = "_QC"
"""Variable name suffix for flag variables"""
FILE_EXTENSION = "nc"
"""filenmae extention for all netcdf files"""

EPOCH = datetime.datetime(1980, 1, 1, 0, 0, 0)
"""dateime referenced in the units of time variables in netCDF files: 1980-01-01"""

STATIC_PARAMETERS_PER_CAST = (
    "EXPOCODE",
    "SECT_ID",
    "STNNBR",
    "CASTNO",
    "_DATETIME",
    "LATITUDE",
    "LONGITUDE",
    "DEPTH",
    "BTLNBR",
    "SAMPNO",
    # extra from porting
    ["DATE", "TIME"],
    "DATE",
    "TIME",
)
"""List of WHP names that are ignored when calling :func:`create_and_fill_data_variables`"""

NON_FLOAT_PARAMETERS = ("CTDNOBS",)
"""params not in :attr:`STATIC_PARAMETERS_PER_CAST` that are also ignored by :func:`create_and_fill_data_variables`"""

UNKNOWN = "UNKNOWN"
"""Value used when some string value isn't found

This is mmostly mitigated by the guarantees of the new CF format, but e.g. section id might be missing
"""

UNSPECIFIED_UNITS = "unspecified"
"""Value used when there are no units"""

STRLEN = 40
"""length of char array variables, hardcoded to 40"""


def bytes_via_temp(xr: xr.Dataset) -> bytes:
    with TemporaryDirectory() as td:
        path = td + "/tmp.nc"
        xr.to_netcdf(path, format="NETCDF3_CLASSIC")
        with open(path, "rb") as f:
            return f.read()


# new behavior will always have the input be a datetime.datetime (Or np.datetime64)
# this function needs modification
def strftime_woce_date_time(dt: xr.DataArray):
    """Take an xr.DataArray with time values in it and convert to strings."""
    if dt is None:
        return (None, None)
    if dt.attrs.get("resolution", 0) >= 1:
        return (dt.dt.strftime("%Y%m%d").item(), None)
    return (dt.dt.strftime("%Y%m%d").item(), dt.dt.strftime("%H%M").item())


# utility functions from libcchdo.formats.netcdf


# name change ascii -> _ascii to avoid builtin conflict
def _ascii(x: str) -> str:
    """Force all codepoints into valid ascii range.

    Works by encoding the str into ascii bytes with the replace err param, then decoding the bytes to str again

    :param x: string with any unicode codepoint in it
    :returns: string with all non ascii codepoints replaced with whatever "replace" does in :py:meth:`str.encode`
    """
    return x.encode("ascii", "replace").decode("ascii")


def simplest_str(s) -> str:
    """Give the simplest string representation.

    If a float is almost equivalent to an integer, swap out for the integer.
    """
    # if type(s) is float:
    if isinstance(s, float):
        # if fns.equal_with_epsilon(s, int(s)):
        # replace with equivalent numpy call
        if np.isclose(s, int(s), atol=1e-6):
            s = int(s)
    return str(s)


def _pad_station_cast(x: str) -> str:
    """Pad a station or cast identifier out to 5 characters.

    This is usually for use in a file name.

    :param x: a string to be padded
    :type x: str
    """
    return simplest_str(x).rjust(5, "0")


def get_filename(expocode, station, cast, extension):
    """Generate the filename for COARDS netCDF files.

    Was ported directly from libcchdo and should have the same formatting behavior
    """
    if extension not in ["hy1", "ctd"]:
        log.warning("File extension is not recognized.")
    station = _pad_station_cast(station)
    cast = _pad_station_cast(cast)

    stem = f"{expocode}_{station}_{cast}_{extension}"
    return f"{stem}.{FILE_EXTENSION}"


def minutes_since_epoch(dt: xr.DataArray, epoch, error=-9):
    """Make the time value for netCDF files.

    The custom implimentation in libcchdo was discarded in favor of the date2num function from cftime.
    Not sure if cftime exsited in the netCDF4 python library at the time.
    """
    return date2num(
        dt.values.astype("<M8[ms]").astype(datetime.datetime),
        epoch,
        calendar="proleptic_gregorian",
    )


# end utility


def get_coards_global_attributes(ds: xr.Dataset, *, profile_type: Literal["B", "C"]):
    """Makes the global attributes of a WHP COARDS netCDF File.

    The order of the attributes is important/fixed, same with case"""

    data_types = {
        "B": "WOCE Bottle",
        "C": "WOCE CTD",
    }

    attrs = {
        "EXPOCODE": ds.expocode.item(),
        "Conventions": "COARDS/WOCE",
        "WOCE_VERSION": "3.0",
    }

    if woce_id := ds.get("section_id"):
        attrs["WOCE_ID"] = woce_id.item()

    attrs["DATA_TYPE"] = data_types[profile_type]
    attrs["STATION_NUMBER"] = ds.station.item()
    attrs["CAST_NUMBER"] = ds["cast"].astype(str).item()

    if bottom_depth := ds.get("btm_depth"):
        attrs["BOTTOM_DEPTH_METERS"] = int(bottom_depth.fillna(FILL_VALUE).item())
    else:
        attrs["BOTTOM_DEPTH_METERS"] = FILL_VALUE

    attrs["Creation_Time"] = datetime.datetime.now(tz=datetime.UTC).strftime(
        "%Y-%m-%dT%H:%M:%S.%fZ"
    )

    _comments = ds.attrs.get("comments", "").splitlines()
    if len(_comments) == 0:
        _comments = [""]
    og_header = "\n".join([_comments[0], *[f"#{line}" for line in _comments[1:]], ""])
    attrs["ORIGINAL_HEADER"] = og_header

    if profile_type == "B":
        ds = ds.stack(ex=("N_PROF", "N_LEVELS"))
        ds = ds.isel(ex=(ds.sample != ""))
        bottle_column = ds.get("bottle_number", ds["sample"])
        attrs["BOTTLE_NUMBERS"] = " ".join(map(simplest_str, bottle_column.values))

        if (btl_quality_codes := ds.get("bottle_number_qc")) is not None:
            attrs["BOTTLE_QUALITY_CODES"] = (
                btl_quality_codes.fillna(9).to_numpy().astype(np.int16)
            )

        attrs["WOCE_BOTTLE_FLAG_DESCRIPTION"] = woce.BOTTLE_FLAG_DESCRIPTION
        attrs["WOCE_WATER_SAMPLE_FLAG_DESCRIPTION"] = woce.WATER_SAMPLE_FLAG_DESCRIPTION

    if profile_type == "C":
        attrs["WOCE_CTD_FLAG_DESCRIPTION"] = woce.CTD_FLAG_DESCRIPTION

    return attrs


def get_dataarrays(ds: xr.Dataset):
    dataarrays = {}
    for whpname, variable in ds.cchdo.to_whp_columns(compact=True).items():
        attrs = {}

        parameter_name = whpname.whp_name

        parameter_unit = whpname.whp_unit or ""
        parameter_key = f"{parameter_name} [{parameter_unit}]"

        # Hacks to match behavior of previous
        if parameter_name == "INSTRUMENT_ID":
            continue

        if parameter_name in STATIC_PARAMETERS_PER_CAST:
            continue

        if parameter_name in NON_FLOAT_PARAMETERS:
            continue

        # porting note: this logic is to match the libcchdo COARDS netcdf varnames
        parameter = PARAMS.get(parameter_key, {})
        _pname = parameter.get("name_netcdf")
        if not _pname:
            log.debug(
                "No netcdf name for %s. Using mnemonic %s.",
                parameter_key,
                parameter_name,
            )
            _pname = parameter_name
        if not _pname:
            raise AttributeError(f"No name found for {parameter_name}")
        pname = _ascii(_pname)

        # porting note: this was taken from the orig libcchdo
        # XXX HACK
        if pname == "oxygen1":
            pname = "oxygen"

        attrs["long_name"] = pname
        if pname == "pressure":
            attrs["positive"] = "down"

        units = UNSPECIFIED_UNITS
        if parameter.get("units_name"):
            units = parameter["units_name"]
        elif parameter_unit != "":
            units = parameter_unit
        attrs["units"] = _ascii(units)

        if parameter_name == "BTL_DATE":
            data = variable.dt.strftime("%Y%m%d").astype(float).to_numpy()
        elif parameter_name == "BTL_TIME":
            data = variable.dt.strftime("%H%M").astype(float).to_numpy()
        elif parameter_name == "CTDETIME" and variable.dtype.char == "m":
            nat_mask = np.isnat(variable)
            scale_factor = np.array(1, dtype=variable.dtype) / np.array(
                1, dtype="timedelta64[s]"
            )
            data = variable.values.astype("float64") * scale_factor
            data[nat_mask] = np.nan
        else:
            data = variable.to_numpy()

        # porting note: hack for new string params that would crash the old converter anyway
        if data.dtype.kind in "OSU":
            data = np.full_like(data, np.nan, dtype=float)

        if data.dtype.kind in "iuf":
            if np.all(np.isnan(data)):
                # the proper fix here would be to make a type dict that knows what is going on with each key
                attrs["data_min"] = float("-inf")  # type: ignore[assignment]
                attrs["data_max"] = float("inf")  # type: ignore[assignment]
            else:
                attrs["data_min"] = np.nanmin(data)
                attrs["data_max"] = np.nanmax(data)

        if parameter.get("format"):
            attrs["C_format"] = _ascii(parameter["format"])
        else:
            # TODO TEST this
            log.debug("Parameter %s has no format. defaulting to '%%f'", parameter_name)
            attrs["C_format"] = "%f"
        if attrs["C_format"].endswith("s"):
            log.debug(
                "Parameter %s does not have a format string acceptable for "
                "numeric data. Defaulting to '%%f' to prevent ncdump "
                "segfault.",
                parameter_name,
            )
            attrs["C_format"] = "%f"

        attrs["WHPO_Variable_Name"] = parameter_name

        dataarrays[pname] = xr.DataArray(
            data, dims=["pressure"], name=pname, attrs=attrs
        )
        dataarrays[pname].encoding["_FillValue"] = None

        if (qc_variable := variable.attrs.get(acc.FLAG_NAME)) is not None:
            qc_name = pname + QC_SUFFIX
            dataarrays[pname].attrs["OBS_QC_VARIABLE"] = qc_name

            qc_attrs = {
                "long_name": qc_name + "_flag",
                "units": "woce_flags",
                "C_format": "%1d",
            }
            qc_data = np.nan_to_num(qc_variable.to_numpy(), nan=9).astype("int16")
            qc_dataarray = xr.DataArray(
                qc_data, dims=["pressure"], name=qc_name, attrs=qc_attrs
            )
            dataarrays[qc_name] = qc_dataarray

    return dataarrays


def get_common_variables(ds: xr.Dataset):
    # In the origional it creates these all manually (not in a loop)
    common_variables = {}
    time_units = f"minutes since {EPOCH.isoformat(' ')}"
    time_data = int(minutes_since_epoch(ds.time, time_units).item())
    common_variables["time"] = xr.DataArray(
        [time_data],
        dims=("time"),
        attrs={
            "long_name": "time",
            "units": time_units,
            "data_min": time_data,
            "data_max": time_data,
            "C_format": "%10d",
        },
    )

    latitude = float(ds.latitude.item())
    common_variables["latitude"] = xr.DataArray(
        np.array([latitude], dtype="float32"),
        dims=["latitude"],
        attrs={
            "long_name": "latitude",
            "units": "degrees_N",
            "data_min": latitude,
            "data_max": latitude,
            "C_format": "%9.4f",
        },
    )
    common_variables["latitude"].encoding["_FillValue"] = None

    longitude = float(ds.longitude.item())
    common_variables["longitude"] = xr.DataArray(
        np.array([longitude], dtype="float32"),
        dims=["longitude"],
        attrs={
            "long_name": "longitude",
            "units": "degrees_E",
            "data_min": longitude,
            "data_max": longitude,
            "C_format": "%9.4f",
        },
    )
    common_variables["longitude"].encoding["_FillValue"] = None

    strs_woce_datetime = strftime_woce_date_time(ds.time)

    date = int(strs_woce_datetime[0] or -9)
    common_variables["woce_date"] = xr.DataArray(
        [date],
        dims=["time"],
        attrs={
            "long_name": "WOCE date",
            "units": "yyyymmdd UTC",
            "data_min": date,
            "data_max": date,
            "C_format": "%8d",
        },
    )

    if strs_woce_datetime[1]:
        time = int(strs_woce_datetime[1] or -9)
        common_variables["woce_time"] = xr.DataArray(
            np.array([time], dtype="int16"),
            dims=["time"],
            attrs={
                "long_name": "WOCE time",
                "units": "hhmm UTC",
                "data_min": time,
                "data_max": time,
                "C_format": "%4d",
            },
        )

    station = simplest_str(ds.station.item()).ljust(STRLEN)[:STRLEN]
    common_variables["station"] = xr.DataArray(
        station,
        attrs={"long_name": "STATION", "units": UNSPECIFIED_UNITS, "C_format": "%s"},
    )
    common_variables["station"].encoding["char_dim_name"] = "string_dimension"

    station = simplest_str(ds.cast.item()).ljust(STRLEN)[:STRLEN]
    common_variables["cast"] = xr.DataArray(
        station,
        attrs={"long_name": "CAST", "units": UNSPECIFIED_UNITS, "C_format": "%s"},
    )
    common_variables["cast"].encoding["char_dim_name"] = "string_dimension"

    return common_variables


def write_bottle(ds: xr.Dataset) -> bytes:
    attrs = get_coards_global_attributes(ds, profile_type="B")
    data_vars = get_dataarrays(ds)
    common_vars = get_common_variables(ds)

    nc_file = xr.Dataset(data_vars={**data_vars, **common_vars}, attrs=attrs)
    return bytes_via_temp(nc_file)


def write_ctd(ds: xr.Dataset) -> bytes:
    attrs = get_coards_global_attributes(ds, profile_type="C")
    data_vars = get_dataarrays(ds)

    if (ctd_nobs := ds.get("ctd_number_of_observations")) is not None:
        stacked_ctd_nobs = ctd_nobs.stack(ex=("N_PROF", "N_LEVELS"))
        valid_ctd_nobs = stacked_ctd_nobs.isel(ex=(stacked_ctd_nobs.sample != ""))
        nobs_data = np.nan_to_num(valid_ctd_nobs.to_numpy(), nan=-999).astype("int64")
        data_vars["number_observations"] = xr.DataArray(
            nobs_data,
            dims=["pressure"],
            attrs={
                "long_name": "number_observations",
                "units": "integer",
                "data_min": np.min(nobs_data),
                "data_max": np.max(nobs_data),
                "C_format": "%1d",
            },
        )
        data_vars["number_observations"].encoding["_FillValue"] = -999
    common_vars = get_common_variables(ds)

    nc_file = xr.Dataset(data_vars={**data_vars, **common_vars}, attrs=attrs)
    return bytes_via_temp(nc_file)


def to_coards(ds: xr.Dataset) -> bytes:
    """Convert an xr.Dataset to a zipfile with COARDS netCDF files inside.

    This function does support mixed CTD and Bottle datasets and will convert using profile_type var on a per profile basis.


    :param ds: A dataset conforming to CCHDO CF/netCDF
    :returns: a zipfile with one or more COARDS netCDF files as members.
    """
    output_files = {}
    for _, profile in ds.groupby("N_PROF", squeeze=False):
        if profile.profile_type.item() == "C":
            data = write_ctd(profile)
            extension = "ctd"
        elif profile.profile_type.item() == "B":
            extension = "hy1"
            data = write_bottle(profile)
        else:
            raise NotImplementedError()

        filename = get_filename(
            profile.expocode.item(),
            profile.station.item(),
            profile.cast.item(),
            extension=extension,
        )
        output_files[filename] = data

    output_zip = BytesIO()
    with ZipFile(output_zip, "w", compression=ZIP_DEFLATED) as zipfile:
        for fname, data in output_files.items():
            zipfile.writestr(fname, data)

    output_zip.seek(0)
    return output_zip.read()



"""A Collection of Flag Schemes."""

from enum import IntEnum


class ExchangeFlag(IntEnum):
    def __init__(self, flag) -> None:
        self.flag = flag

    @property
    def definition(self):
        return self._flag_definitions[self.flag]

    @property
    def cf_def(self):
        return "_".join(self.definition.lower().replace(".", "").split())

    @property
    def has_value(self):
        if self.flag in self._no_data_flags:
            return False
        return True


class ExchangeBottleFlag(ExchangeFlag):
    """Enum representing a WHP Bottle flag.

    This flag represents information about the sampling device itself
    (i.e. the niskin bottle). It should only be used for "BTLNBR_FLAG_W"
    values and should never be used with CTD files.
    """

    NOFLAG = 0  # no idea if this will cause issue
    NO_INFO = 1
    GOOD = 2
    LEAKING = 3
    BAD_TRIP = 4
    NOT_REPORTED = 5
    DISCREPANCY = 6
    UNKNOWN = 7
    PAIR = 8
    NOT_SAMPLED = 9

    @property
    def _no_data_flags(self):
        return (1, 5, 9)

    @property
    def _flag_definitions(self):
        return {
            0: "No Flag assigned",
            1: "Bottle information unavailable.",
            2: "No problems noted.",
            3: "Leaking.",
            4: "Did not trip correctly.",
            5: "Not reported.",
            6: "Significant discrepancy in measured values between Gerard and Niskin bottles.",
            7: "Unknown problem.",
            8: "Pair did not trip correctly. Note that the Niskin bottle can trip at an unplanned depth while the Gerard trips correctly and vice versa.",
            9: "Samples not drawn from this bottle.",
        }


class ExchangeSampleFlag(ExchangeFlag):
    NOFLAG = 0  # no idea if this will cause issue
    MISSING = 1
    GOOD = 2
    QUESTIONABLE = 3
    BAD = 4
    NOT_REPORTED = 5
    MEAN = 6
    CHROMA_MANUAL = 7
    CHROMA_IRREGULAR = 8
    NOT_SAMPLED = 9

    @property
    def _no_data_flags(self):
        return (1, 5, 9)

    @property
    def _flag_definitions(self):
        return {
            0: "No Flag assigned",
            1: "Sample for this measurement was drawn from water bottle but analysis not received.",
            2: "Acceptable measurement.",
            3: "Questionable measurement.",
            4: "Bad measurement.",
            5: "Not reported.",
            6: "Mean of replicate measurements",
            7: "Manual chromatographic peak measurement.",
            8: "Irregular digital chromatographic peak integration.",
            9: "Sample not drawn for this measurement from this bottle.",
        }


class ExchangeCTDFlag(ExchangeFlag):
    NOFLAG = 0  # no idea if this will cause issue
    UNCALIBRATED = 1
    GOOD = 2
    QUESTIONABLE = 3
    BAD = 4
    NOT_REPORTED = 5
    INTERPOLATED = 6
    DESPIKED = 7
    NOT_SAMPLED = 9

    @property
    def _no_data_flags(self):
        return (5, 9)

    @property
    def _flag_definitions(self):
        return {
            0: "No Flag assigned",
            1: "Not calibrated.",
            2: "Acceptable measurement.",
            3: "Questionable measurement.",
            4: "Bad measurement.",
            5: "Not reported.",
            6: "Interpolated over a pressure interval larger than 2 dbar.",
            7: "Despiked.",
            9: "Not sampled.",
        }



from csv import reader as csv_reader

# TODO: switch to files().joinpath().open when python 3.8 is dropped
# 2023-04-16
from importlib.resources import open_text
from io import BytesIO
from itertools import zip_longest
from zipfile import ZIP_DEFLATED, ZipFile

import numpy as np
import xarray as xr

from cchdo.hydro import accessors as acc

#
CTD_ZIP_FILE_EXTENSION = "ct.zip"
CTD_FILE_EXTENSION = "ct.txt"
BOTTLE_FILE_EXTENSION = "hy.txt"

FILL_VALUE = -9

ASTERISK_FLAG = "*" * 7

CHARACTER_PARAMETERS = ["STNNBR", "SAMPNO", "BTLNBR"]

COLUMN_WIDTH = 8
SAFE_COLUMN_WIDTH = COLUMN_WIDTH - 1

UNKNONW_TIME_FILL = "0000"

BOTTLE_FLAGS = {
    1: "Bottle information unavailable.",
    2: "No problems noted.",
    3: "Leaking.",
    4: "Did not trip correctly.",
    5: "Not reported.",
    6: (
        "Significant discrepancy in measured values between Gerard and Niskin bottles."
    ),
    7: "Unknown problem.",
    8: (
        "Pair did not trip correctly. Note that the Niskin bottle can trip at "
        "an unplanned depth while the Gerard trips correctly and vice versa."
    ),
    9: "Samples not drawn from this bottle.",
}

CTD_FLAGS = {
    1: "Not calibrated",
    2: "Acceptable measurement",
    3: "Questionable measurement",
    4: "Bad measurement",
    5: "Not reported",
    6: "Interpolated over >2 dbar interval",
    7: "Despiked",
    8: "Not assigned for CTD data",
    9: "Not sampled",
}


WATER_SAMPLE_FLAGS = {
    1: (
        "Sample for this measurement was drawn from water bottle but analysis "
        "not received."
    ),
    2: "Acceptable measurement.",
    3: "Questionable measurement.",
    4: "Bad measurement.",
    5: "Not reported.",
    6: "Mean of replicate measurements.",
    7: "Manual chromatographic peak measurement.",
    8: "Irregular digital chromatographic peak integration.",
    9: "Sample not drawn for this measurement from this bottle.",
}


def flag_description(flag_map):
    return ":".join(
        [":"] + [f"{flag} = {value}" for flag, value in flag_map.items()] + ["\n"]
    )


BOTTLE_FLAG_DESCRIPTION = flag_description(BOTTLE_FLAGS)

CTD_FLAG_DESCRIPTION = flag_description(CTD_FLAGS)

WATER_SAMPLE_FLAG_DESCRIPTION = flag_description(WATER_SAMPLE_FLAGS)

_UNWRITTEN_COLUMNS = [
    "EXPOCODE",
    "SECT_ID",
    "LATITUDE",
    "LONGITUDE",
    "DEPTH",
    "_DATETIME",
]

# machinery from COARDS


def simplest_str(s) -> str:
    """Give the simplest string representation.

    If a float is almost equivalent to an integer, swap out for the integer.
    """
    # if type(s) is float:
    if isinstance(s, float):
        # if fns.equal_with_epsilon(s, int(s)):
        # replace with equivalent numpy call
        if np.isclose(s, int(s), atol=1e-6):
            s = int(s)
    return str(s)


def _pad_station_cast(x: str) -> str:
    """Pad a station or cast identifier out to 5 characters. This is usually
    for use in a file name.

    :param x: a string to be padded
    :type x: str
    """
    return simplest_str(x).rjust(5, "0")


def get_filename(expocode, station, cast, file_ext):
    station = _pad_station_cast(station)
    cast = _pad_station_cast(cast)
    return "{}.{}".format(
        f"{expocode}_{station}_{cast}",
        file_ext,
    )


# END machinery


def convert_fortran_format_to_c(ffmt: str):
    """Simplistic conversion from Fortran format string to C format string.

    This only operates on F formats.
    """
    if not ffmt:
        return ffmt
    if ffmt.startswith("F"):
        return f"%{ffmt[1:]}f"
    elif ffmt.startswith("I"):
        return f"%{ffmt[1:]}d"
    elif ffmt.startswith("A"):
        return f"%{ffmt[1:]}s"
    elif "," in ffmt:
        # WOCE specifies things like 1X,A7, so only convert the last bit.
        ffmt = ffmt.split(",")[1]
        return convert_fortran_format_to_c(ffmt)
    return ffmt


def get_exwoce_params():
    """Return a dictionary of WOCE parameters allowed for Exchange conversion.

    :return: {'PMNEMON': {'unit_mnemonic': 'WOCE', 'range': [0.0, 10.0], 'format': '%8.3f'}}
    """
    with open_text(
        "cchdo.hydro.legacy.woce", "woce_params_for_exchange_to_woce.csv"
    ) as params:
        reader = csv_reader(params)
        params = {}
        for order, row in enumerate(reader):
            # First line is header
            if order == 0:
                continue

            if row[-1] == "x":
                continue
            if not row[1]:
                row[1] = None
            if row[2]:
                prange = list(map(float, row[2].split(",")))
            else:
                prange = None
            if not row[3]:
                row[3] = None
            params[row[0]] = {
                "unit_mnemonic": row[1],
                "range": prange,
                "format": convert_fortran_format_to_c(row[3]),
                "order": order,
            }
        return params


_EXWOCE_PARAMS = get_exwoce_params()


def writeable_columns(ds: xr.Dataset, is_ctd=False):
    """Return the columns that belong in a WOCE data file."""
    CTD_IGNORE = ["STNNBR", "CASTNO", "SAMPNO"]

    # Filter with whitelist and rewrite format strings to WOCE standard.
    whitelisted_columns = []
    for param, col in ds.cchdo.to_whp_columns(compact=True).items():
        key = param.whp_name
        if key in _UNWRITTEN_COLUMNS:
            continue
        if is_ctd and key in CTD_IGNORE:
            continue
        if key not in _EXWOCE_PARAMS:
            continue
        info = _EXWOCE_PARAMS[key]
        fmt = info["format"]
        if fmt:
            col.attrs["cchdo.hydro._format"] = fmt
        col.attrs["cchdo.hydro._display_order"] = info["order"]
        whitelisted_columns.append(col)
    return sorted(
        whitelisted_columns, key=lambda col: col.attrs["cchdo.hydro._display_order"]
    )


def columns_and_base_format(dfile, is_ctd=False):
    """Return columns and base format for WOCE fixed column data."""
    columns = writeable_columns(dfile, is_ctd=is_ctd)
    num_qualt = len(list(filter(lambda col: acc.FLAG_NAME in col.attrs, columns)))
    col_format = "{{{0}:>8}}"
    base_format = "".join([col_format.format(iii) for iii in range(len(columns))])
    qualt_colsize = max((len("QUALT#"), num_qualt))
    qualt_format = f"{{}}:>{qualt_colsize}"
    base_format += " {" + qualt_format.format(len(columns)) + "}\n"
    return columns, base_format


def truncate_row(lll):
    """Return a new row where all items are less than or equal to column width.

    Warnings will be given for any truncations.
    """
    truncated = []
    for xxx in lll:
        if len(xxx) > COLUMN_WIDTH:
            trunc = xxx[:COLUMN_WIDTH]
            # jlog.warn(u'Truncated {0!r} to {1!r} because longer than {2} '
            # j         'characters.'.format(xxx, trunc, COLUMN_WIDTH))
            xxx = trunc
        truncated.append(xxx)
    return truncated


def write_data(ds, columns, base_format):
    """Write WOCE data in fixed width columns.

    columns and base_format should be obtained from
    columns_and_base_format()
    """

    def parameter_name_of(column):
        return column.attrs["whp_name"]

    def units_of(column):
        if "whp_unit" in column.attrs:
            return column.attrs["whp_unit"]
        else:
            return ""

    def quality_flags_of(column):
        return ASTERISK_FLAG if acc.FLAG_NAME in column.attrs else ""

    all_headers = list(map(parameter_name_of, columns))
    all_units = list(map(units_of, columns))
    all_asters = list(map(quality_flags_of, columns))

    all_headers.append("QUALT1")
    all_units.append("*")
    all_asters.append("*")

    record2 = base_format.format(*truncate_row(all_headers))
    record3 = base_format.format(*truncate_row(all_units))
    record4 = base_format.format(*truncate_row(all_asters))

    data_lines = []

    data = []
    flags = []
    for column in columns:
        format_str = column.attrs.get("cchdo.hydro._format", "%s")
        str_column = [format_str % x for x in column.fillna(FILL_VALUE).to_numpy()]
        for i, d in enumerate(str_column):
            if len(d) > COLUMN_WIDTH:
                extra = len(d) - COLUMN_WIDTH
                leading_extra = d[:extra]
                if len(leading_extra.strip()) == 0:
                    d = d[extra:]
                else:
                    d = d[:-extra]

                str_column[i] = d
        data.append(str_column)

        if acc.FLAG_NAME in column.attrs:
            flags.append(
                [
                    f"{x:1d}"
                    for x in column.attrs[acc.FLAG_NAME]
                    .fillna(9)
                    .to_numpy()
                    .astype(int)
                ]
            )

    for row_d, row_f in zip_longest(
        zip(*data, strict=True), zip(*flags, strict=True), fillvalue=""
    ):
        data_lines.append(base_format.format(*row_d, "".join(row_f)))

    return "".join([record2, record3, record4, *data_lines])


def write_bottle(ds: xr.Dataset):
    """How to write a Bottle WOCE file."""
    # Look through datetime for begin and end dates
    begin_date = np.min(ds.time).dt.strftime("%Y%m%d").values
    end_date = np.max(ds.time).dt.strftime("%Y%m%d").values

    expocodes = "/".join(np.unique(ds.expocode))
    sect_id = "NONE"
    if "section_id" in ds:
        sect_id = "/".join(np.unique(ds.section_id))

    columns, base_format = columns_and_base_format(ds)

    vals = [""] * (len(columns) + 1)
    empty_line = base_format.format(*vals)
    record_len = len(empty_line) - 2

    record_1 = f"EXPOCODE {expocodes:s} WHP-ID {sect_id:s} CRUISE DATES {begin_date} TO {end_date}"

    record_1 += " " * (record_len - len(record_1))
    record_1 += "*"
    record_1 += "\n"

    data = write_data(ds, columns, base_format)
    return f"{record_1}{data}".encode("ascii", "replace")


def write_ctd(ds: xr.Dataset):
    """How to write a CTD WOCE file."""
    # We can only write the CTD file if there is a unique
    # EXPOCODE, STNNBR, and CASTNO in the file.
    if ds.sizes["N_PROF"] != 1:
        raise NotImplementedError("can only write single profile")

    expocode = "/".join(np.unique(ds.expocode))
    section = "NONE"
    if "section_id" in ds:
        section = "/".join(np.unique(ds.section_id))
    station = ds.station[0]
    cast = ds.cast[0]

    columns, base_format = columns_and_base_format(ds, is_ctd=True)

    date = ds.time.dt.strftime("%m%d%y")[0].item()

    record1 = f"EXPOCODE {expocode} WHP-ID {section} DATE {date}"
    # 2 at end of line denotes record 2
    record2 = f"STNNBR {station: >8s} CASTNO {cast: >3d} NO. RECORDS={ds.sizes['N_LEVELS']: >5d}"
    # 3 denotes record 3
    instrument_no = ds.get("instrument_id", ["-9"])[0]
    sampling_rate = ds.get("ctd_sampling_rate", [-9])[0]
    record3 = (
        f"INSTRUMENT NO. {instrument_no: >5s} SAMPLING RATE {sampling_rate:>6.2f} HZ"
    )

    headers = f"{record1}\n{record2}\n{record3}"
    data = write_data(ds, columns, base_format)
    return f"{headers}\n{data}".encode("ascii", "replace")


def to_woce(ds: xr.Dataset) -> bytes:
    output_files = {}
    profile_type_nd = np.unique(ds.profile_type)

    if len(profile_type_nd) != 1:
        raise NotImplementedError("Cannot convert mixed profile types to woce")
    profile_type = profile_type_nd.item()

    if profile_type == "B":
        return write_bottle(ds)

    elif profile_type == "C":
        for _, profile in ds.groupby("N_PROF", squeeze=False):
            compact = profile.cchdo.compact_profile()
            data = write_ctd(compact)

            filename = get_filename(
                profile.expocode.item(),
                profile.station.item(),
                profile.cast.item(),
                file_ext=CTD_FILE_EXTENSION,
            )
            output_files[filename] = data

        output_zip = BytesIO()
        with ZipFile(output_zip, "w", compression=ZIP_DEFLATED) as zipfile:
            for fname, data in output_files.items():
                zipfile.writestr(fname, data)

        output_zip.seek(0)
        return output_zip.read()
    else:
        raise NotImplementedError("Unknown profile_type")



import io
from importlib.resources import open_binary

import pytest

from cchdo.hydro.exchange import read_csv, read_exchange
from cchdo.hydro.exchange.helpers import simple_bottle_exchange


@pytest.fixture
def nc_empty():
    return read_exchange(io.BytesIO(simple_bottle_exchange()))


@pytest.fixture
def nc_placeholder():
    params = ("OXYGEN", "OXYGEN_FLAG_W", "DELC14", "DELC14_FLAG_W", "C14ERR")
    units = ("UMOL/KG", "", "/MILLE", "", "/MILLE")
    data = ("-999", "1", "-999", "1", "-999")
    return read_exchange(
        io.BytesIO(simple_bottle_exchange(params=params, units=units, data=data))
    )


@pytest.fixture(scope="function")
def nc_placeholders():
    # Multiple empty slots for this one
    with open_binary("cchdo.hydro.tests.data", "merge_placeholders.csv") as f:
        return read_csv(f)



import io
import os
from collections.abc import Mapping
from zipfile import ZipFile

import requests

from . import _hydro_appdirs

bottle_uri = "https://cchdo.ucsd.edu/search?q=a&download=exchange%2cbottle"
bottle_fname = "bottle_data.zip"


def _cache_dir():
    path = _hydro_appdirs.user_cache_dir
    os.makedirs(path, exist_ok=True)
    return path


def load_cchdo_bottle_data():
    """Downloads some CCHDO data for playing with..."""
    path = os.path.join(_cache_dir(), bottle_fname)
    with requests.get(bottle_uri, stream=True) as r:
        r.raise_for_status()
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)


class CCHDOBottleData(Mapping):
    def __init__(self):
        self.path = os.path.join(_hydro_appdirs.user_cache_dir, bottle_fname)
        try:
            with ZipFile(self.path) as f:
                self.files = f.namelist()
        except FileNotFoundError:
            load_cchdo_bottle_data()
            with ZipFile(self.path) as f:
                self.files = f.namelist()

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        yield from self.files

    def __getitem__(self, key):
        if key not in self.files:
            raise KeyError()
        with ZipFile(self.path) as f:
            return io.BytesIO(f.read(key))



import xarray as xr
from js import Uint8Array
from pyscript import display

from cchdo.hydro import accessors  # noqa: F401


def logger(msg):
    display(msg, target="log", append=True)


def load_netcdf(array_buffer):
    with open("out.nc", "wb") as f:
        f.write(bytearray(Uint8Array.new(array_buffer)))
    return xr.load_dataset("out.nc")


def make_derived(array_buffer, type):
    logger(type)
    xr = load_netcdf(array_buffer)
    if type == "to_sum":
        return xr.cchdo.to_sum(), "summary.txt"
    if type == "to_woce":
        return xr.cchdo.to_woce(), xr.cchdo.gen_fname("woce")
    if type == "to_coards":
        return xr.cchdo.to_coards(), xr.cchdo.gen_fname("coards")


__export__ = [
    "make_derived",
]



import numpy as np

from cchdo.params import WHPNames

DIMS = ("N_PROF", "N_LEVELS")

GEOMETRY_VARS = ("expocode", "station", "cast", "section_id", "time")

FILLS_MAP = {"string": "", "integer": np.nan, "decimal": np.nan}

EXPOCODE = WHPNames["EXPOCODE"]
STNNBR = WHPNames["STNNBR"]
CASTNO = WHPNames["CASTNO"]
SAMPNO = WHPNames["SAMPNO"]
DATE = WHPNames["DATE"]
TIME = WHPNames["TIME"]
LATITUDE = WHPNames["LATITUDE"]
LONGITUDE = WHPNames["LONGITUDE"]
CTDPRS = WHPNames[("CTDPRS", "DBAR")]
BTLNBR = WHPNames["BTLNBR"]

COORDS = [
    EXPOCODE,
    STNNBR,
    CASTNO,
    SAMPNO,
    DATE,
    TIME,
    LATITUDE,
    LONGITUDE,
    CTDPRS,
]



import io

from cchdo.params import WHPNames


def simple_bottle_exchange(
    params=None, units=None, data=None, comments: str | None = None
):
    stamp = "BOTTLE,test"
    min_params = [
        "EXPOCODE",
        "STNNBR",
        "CASTNO",
        "SAMPNO",
        "LATITUDE",
        "LONGITUDE",
        "DATE",
        "TIME",
        "CTDPRS",
    ]
    min_units = ["", "", "", "", "", "", "", "", "DBAR"]
    min_line = ["TEST", "1", "1", "1", "0", "0", "20200101", "0000", "0"]
    end = "END_DATA"

    if params is not None:
        min_params.extend(params)
    if units is not None:
        min_units.extend(units)
    if data is not None:
        min_line.extend(data)

    if comments is not None:
        comments = "\n".join([f"#{line}" for line in comments.splitlines()])
        simple = "\n".join(
            [
                stamp,
                comments,
                ",".join(min_params),
                ",".join(min_units),
                ",".join(min_line),
                end,
            ]
        )
    else:
        simple = "\n".join(
            [stamp, ",".join(min_params), ",".join(min_units), ",".join(min_line), end]
        )
    return simple.encode("utf8")


def gen_template(
    ftype="B",
    param_counts: dict[str, int] | None = None,
    min_count=5,
    filter_erddap=False,
):
    from . import FileType, read_csv

    ftype = FileType(ftype)

    exclude = {
        "EXPOCODE",
        "STNNBR",
        "CASTNO",
        "SAMPNO",
        "LATITUDE",
        "LONGITUDE",
        "DATE",
        "TIME",
        "CTDPRS",
        "BTL_TIME",
        "BTL_DATE",
    }

    params = [
        "EXPOCODE",
        "STNNBR",
        "CASTNO",
        "SAMPNO",
        "LATITUDE",
        "LONGITUDE",
        "DATE",
        "TIME",
        "CTDPRS [DBAR]",
    ]
    data = ["TEST", "1", "1", "1", "0", "0", "20200101", "0000", "0"]
    for name in set(WHPNames.values()):
        if filter_erddap and not name.in_erddap:
            continue
        if name.whp_name in exclude:
            continue
        if param_counts is not None:
            if param_counts.get(name.nc_name, 0) < min_count:
                continue
        if ftype == FileType.CTD and name.flag_w in {
            "woce_discrete",
            "woce_bottle",
            "no_flags",
        }:
            continue
        if name.whp_unit is not None:
            param_name = f"{name.whp_name} [{name.whp_unit}]"
        else:
            param_name = name.whp_name
        params.append(param_name)

        data.append("-999")
        if name.flag_w is not None and name.flag_w != "no_flags":
            params.append(f"{param_name}_FLAG_W")
            data.append("9")

        if name.error_name is not None:
            if name.whp_unit is not None:
                params.append(f"{name.error_name} [{name.whp_unit}]")
            else:
                params.append(name.error_name)
            data.append("-999")

    # special case
    params.extend(["BTL_DATE", "BTL_TIME"])
    data.extend(["20220421", "0944"])

    file = f"{','.join(params)}\n{','.join(data)}"
    data_file = io.BytesIO(file.encode("utf8"))

    return read_csv(data_file, ftype=ftype)



from cchdo.params import WHPNames

print(
    """.. list-table:: Variables In ERDDAP
   :header-rows: 1

   * - netcdf varname
     - Units
     - In ERDDAP"""
)

seen = set()
for name in sorted(WHPNames.values()):
    if name in seen:
        continue
    seen.add(name)
    yes = ":bdg-success:`yes`"
    no = ":bdg-danger:`no`"

    print(
        f"""   * - ``{name.nc_name}``
     - {name.cf_unit}
     - {yes if name.in_erddap else no}"""
    )



from enum import StrEnum, auto
from typing import Literal

FileTypes = Literal["C", "B"]


class FileType(StrEnum):
    CTD = "C"
    BOTTLE = "B"


FileTypeType = FileTypes | FileType

PrecisionSources = Literal["file", "database"]


class PrecisionSource(StrEnum):
    FILE = auto()
    DATABASE = auto()


PrecisionSourceType = PrecisionSources | PrecisionSource

WHPNameAttr = str | list[str]



import os
import warnings
from base64 import b64encode

import xarray as xr

from cchdo.hydro import accessors  # noqa: F401
from cchdo.hydro.exchange import read_exchange
from cchdo.hydro.exchange.exceptions import (
    ExchangeDataFlagPairError,
    ExchangeParameterUndefError,
)
from cchdo.params import WHPNames


def make_netcdf_file_json(path):
    with open(path, "rb") as f:
        return {
            "name": "netcdf.nc",  # not important isn't used
            "type": "application/x-netcdf",
            "body": b64encode(f.read()).decode("utf8"),
        }


def p_file(file_m):
    t_dir, file, file_metadata, roundtrip = file_m
    checks = {"flags": False}
    unknown_params = []

    warnings.simplefilter("ignore")
    if ("OSNUM", None) not in WHPNames:
        WHPNames.add_alias(("OSNUM", None), ("EVENT_NUMBER", None))

    if file_metadata["data_type"] == "ctd" and ("NITRATE", "UMOL/KG") not in WHPNames:
        # HOT names that are a little dangerous to have in the real DB
        WHPNames.add_alias(("NITRATE", "UMOL/KG"), ("CTDNITRATE", "UMOL/KG"))
        WHPNames.add_alias(("CHLPIG", "uG/L"), ("CTDFLUOR", "MG/M^3"))
        WHPNames.add_alias(("CHLPIG", "UG/L"), ("CTDFLUOR", "MG/M^3"))

        # other one off things
        WHPNames.add_alias(("CTDFLUOR", "UG/L_UNCALIBRATED"), ("CTDFLUOR", "MG/M^3"))
        WHPNames.add_alias(("_OS_ID", None), ("EVENT_NUMBER", None))
        WHPNames.add_alias(("CTDSAL", "PPS-78"), ("CTDSAL", "PSS-78"))

    try:
        ex_xr = read_exchange(file, checks=checks)
    except ExchangeParameterUndefError as err:
        return (500, repr(err), file_metadata, err.error_data)
    except ExchangeDataFlagPairError:
        try:
            ex_xr = read_exchange(file, fill_values=("-999", "-99"), checks=checks)
        except (ValueError, KeyError) as err:
            return (500, repr(err), file_metadata, unknown_params)
    except (ValueError, KeyError) as err:
        return (500, repr(err), file_metadata, unknown_params)

    to_path = os.path.join(t_dir, f"{file_metadata['id']}_{ex_xr.cchdo.gen_fname()}")
    ex_xr.to_netcdf(to_path)
    if roundtrip:
        ds = xr.load_dataset(to_path)
        try:
            ex = ds.cchdo.to_exchange()
        except ValueError as err:
            return (
                500,
                f"Roundtrip back to exchange: {err!r}",
                file_metadata,
                unknown_params,
            )
        try:
            ex_xr = read_exchange(ex, checks=checks)
        except ValueError as err:
            return (
                500,
                f"Roundtrip back to and from exchange: {err!r}",
                file_metadata,
                unknown_params,
            )

    return (200, to_path, file_metadata, unknown_params)


def p_file_cf(file_m):
    t_dir, file, file_metadata = file_m

    warnings.simplefilter("ignore")

    ex_xr = xr.load_dataset(file)
    try:
        ex_xr.cchdo.to_exchange()
        exchange_ok = True
    except Exception:
        exchange_ok = False

    try:
        ex_xr.cchdo.to_coards()
        coards_ok = True
    except Exception:
        coards_ok = False

    try:
        ex_xr.cchdo.to_woce()
        woce_ok = True
    except Exception:
        woce_ok = False

    try:
        ex_xr.cchdo.to_sum()
        sum_ok = True
    except Exception:
        sum_ok = False

    return (file_metadata, exchange_ok, coards_ok, woce_ok, sum_ok)



"""Functions that hopefully can migrate from a past version of data to a future version."""

from abc import ABC, abstractmethod

import xarray as xr


class MigrationABC(ABC):
    version_from = "1.0.0.0"

    def can_migrate(self, ds: xr.Dataset) -> bool:
        return self.version_from in ds.attrs.get("software_version", "")

    @abstractmethod
    def migrate(self, ds: xr.Dataset) -> xr.Dataset:
        return ds



import io
from importlib.resources import open_binary

import numpy as np
import pytest

from cchdo.hydro import accessors  # noqa: F401
from cchdo.hydro.exchange import read_exchange
from cchdo.hydro.exchange.exceptions import (
    ExchangeBOMError,
    ExchangeDuplicateParameterError,
    ExchangeEncodingError,
    ExchangeParameterUndefError,
)
from cchdo.hydro.exchange.helpers import simple_bottle_exchange


def test_btl_date_time():
    raw = simple_bottle_exchange(
        params=("BTL_DATE", "BTL_TIME"), units=("", ""), data=("20200101", "1234")
    )
    ex_xr = read_exchange(io.BytesIO(raw))

    assert "bottle_time" in ex_xr.variables
    assert "bottle_date" not in ex_xr.variables
    assert ex_xr["bottle_time"].values == [[np.datetime64("2020-01-01T12:34")]]


def test_btl_date_time_fill_round_trip():
    raw = simple_bottle_exchange(
        params=("BTL_DATE", "BTL_TIME"), units=("", ""), data=("-999", "-999")
    )
    ex_xr = read_exchange(io.BytesIO(raw))

    assert "bottle_time" in ex_xr.variables
    np.testing.assert_array_equal(ex_xr["bottle_time"].values, [[np.datetime64("nat")]])

    # test the round trip
    ex = ex_xr.cchdo.to_exchange()
    assert b"nan" not in ex
    ex_xr_rt = read_exchange(io.BytesIO(ex))
    np.testing.assert_array_equal(
        ex_xr_rt["bottle_time"].values, [[np.datetime64("nat")]]
    )


def test_btl_date_time_missing_warn():
    raw = simple_bottle_exchange(
        params=("BTL_DATE", "BTL_TIME"), units=("", ""), data=("20200101", "34")
    )
    with pytest.warns(UserWarning):
        ex_xr = read_exchange(io.BytesIO(raw))

    assert ex_xr["bottle_time"].values == [[np.datetime64("2020-01-01T00:34")]]


def test_ctd_nan():
    test_data = b"""CTD,test
# some comment
NUMBER_HEADERS = 8
EXPOCODE = TEST
STNNBR = 1
CASTNO = 1
DATE = 20200101
TIME = 0000
LATITUDE = nan
LONGITUDE = 0
CTDPRS
DBAR
0
END_DATA
"""
    with pytest.raises(ValueError):
        read_exchange(test_data)


def test_file_seperator():
    test_data = b"""CTD,test
# some comment
NUMBER_HEADERS = 8
EXPOCODE = TEST
STNNBR = 1
CASTNO = 1
DATE = 20200101
TIME = 0000
LATITUDE = 0
LONGITUDE = 0
CTDPRS
DBAR
0
END_DATA
"""
    ex = read_exchange(test_data + test_data, file_seperator="END_DATA")
    assert ex.sizes["N_PROF"] == 2


@pytest.mark.parametrize(
    "data,error",
    [
        (io.BytesIO("À".encode("latin-1")), ExchangeEncodingError),
        (io.BytesIO("\ufeffBOTTLE".encode()), ExchangeBOMError),
    ],
)
def test_reject_bad_examples(data, error):
    with pytest.raises(error):
        read_exchange(data)


@pytest.mark.parametrize(
    "uri", ["https://cchdo.ucsd.edu/exchange.csv", "http://cchdo.ucsd.edu/exchange.csv"]
)
def test_http_loads(uri, requests_mock):
    requests_mock.get(uri, content=simple_bottle_exchange())
    read_exchange(uri)


@pytest.mark.parametrize("flag", ["1", "2", "3", "4", "6", "7"])
def test_pressure_flags(flag):
    raw = simple_bottle_exchange(params=("CTDPRS_FLAG_W",), units=("",), data=(flag,))
    ex_xr = read_exchange(io.BytesIO(raw))

    assert "pressure_qc" in ex_xr.variables
    assert "pressure_qc" in ex_xr.pressure.attrs["ancillary_variables"]
    assert ex_xr.pressure_qc.data == [int(flag)]
    assert "pressure" in ex_xr.coords


@pytest.mark.parametrize("flag", ["5", "8", "9"])
def test_pressure_flags_bad(flag):
    raw = simple_bottle_exchange(params=("CTDPRS_FLAG_W",), units=("",), data=(flag,))
    with pytest.raises(ValueError):
        read_exchange(io.BytesIO(raw))


def test_duplicate_name_different_units():
    raw = simple_bottle_exchange(
        params=("CTDTMP", "CTDTMP"), units=("ITS-90", "IPTS-68"), data=("-999", "-999")
    )
    ex_xr = read_exchange(io.BytesIO(raw))
    assert "ctd_temperature" in ex_xr.variables
    assert "ctd_temperature_68" in ex_xr.variables


def test_duplicate_name_different_units_keep_flags():
    raw = simple_bottle_exchange(
        params=("CTDTMP", "CTDTMP_FLAG_W", "CTDTMP", "CTDTMP_FLAG_W"),
        units=("ITS-90", "", "IPTS-68", ""),
        data=("-999", "9", "-999", "9"),
    )
    ex_xr = read_exchange(io.BytesIO(raw))
    assert "ctd_temperature" in ex_xr.variables
    assert "ctd_temperature_68" in ex_xr.variables
    assert "ctd_temperature_qc" in ex_xr.variables
    assert "ctd_temperature_68_qc" in ex_xr.variables


def test_duplicate_name_same_units():
    raw = simple_bottle_exchange(
        params=("CTDTMP", "CTDTMP"), units=("ITS-90", "ITS-90"), data=("-999", "-999")
    )
    with pytest.raises(ExchangeDuplicateParameterError):
        read_exchange(io.BytesIO(raw))


def test_multiple_unknown_params():
    raw = simple_bottle_exchange(
        params=("TEST1", "TEST2"), units=("TEST3", "TEST4"), data=("-999", "-999")
    )
    with pytest.raises(ExchangeParameterUndefError) as execinfo:
        read_exchange(io.BytesIO(raw))

    assert hasattr(execinfo.value, "error_data")
    assert execinfo.value.error_data == ["TEST1 [TEST3]", "TEST2 [TEST4]"]


def test_multiple_unknown_params_ignore():
    raw = simple_bottle_exchange(
        params=("TEST1", "TEST2"), units=("TEST3", "TEST4"), data=("-999", "-999")
    )
    read_exchange(io.BytesIO(raw), ignore_columns=("TEST1 [TEST3]", "TEST2 [TEST4]"))


def test_multiple_unknown_params_flag():
    raw = simple_bottle_exchange(
        params=("OXYGEN", "OXYGEN_FLAG_W"), units=("UMOL/KG", ""), data=("-999", "2")
    )
    # The flag input above is intentionally invalid, this should ignore that
    ex = read_exchange(io.BytesIO(raw), ignore_columns=("OXYGEN [UMOL/KG]_FLAG_W",))
    assert "oxygen_qc" not in ex


def test_alternate_params():
    raw = simple_bottle_exchange(
        params=("CTDTMP", "CTDTMP_ALT_1"),
        units=("ITS-90", "ITS-90"),
        data=("-999", "-999"),
    )
    ex_xr = read_exchange(io.BytesIO(raw))
    assert "ctd_temperature" in ex_xr.variables
    assert "ctd_temperature_alt_1" in ex_xr.variables


def test_alternate_params_flags():
    raw = simple_bottle_exchange(
        params=("CTDTMP", "CTDTMP_ALT_1", "CTDTMP_ALT_1_FLAG_W"),
        units=("ITS-90", "ITS-90", ""),
        data=("-999", "-999", "9"),
    )
    ex_xr = read_exchange(io.BytesIO(raw))
    assert "ctd_temperature" in ex_xr.variables
    assert "ctd_temperature_alt_1" in ex_xr.variables
    assert "ctd_temperature_alt_1_qc" in ex_xr.variables
    assert "ctd_temperature_qc" not in ex_xr.variables


def test_fix_bottle_time_span():
    test_data = open_binary("cchdo.hydro.tests.data", "btl_time_span.csv")
    expected = np.array(
        [
            [
                "2012-09-05T01:15:00.000000000",
                "2012-09-05T01:33:00.000000000",
                "2012-09-05T01:32:00.000000000",
                "2012-09-05T01:28:00.000000000",
                "2012-09-05T01:27:00.000000000",
                "2012-09-05T01:26:00.000000000",
                "2012-09-05T01:25:00.000000000",
                "2012-09-05T01:23:00.000000000",
                "2012-09-05T01:22:00.000000000",
                "2012-09-05T01:20:00.000000000",
                "2012-09-05T01:18:00.000000000",
                "2012-09-05T01:16:00.000000000",
                "2012-09-05T01:13:00.000000000",
                "2012-09-05T01:10:00.000000000",
                "2012-09-05T01:08:00.000000000",
                "2012-09-05T01:05:00.000000000",
                "2012-09-05T01:02:00.000000000",
                "2012-09-05T00:59:00.000000000",
                "2012-09-05T00:55:00.000000000",
                "2012-09-05T00:50:00.000000000",
                "2012-09-05T00:45:00.000000000",
                "2012-09-05T00:41:00.000000000",
                "2012-09-05T00:36:00.000000000",
                "2012-09-05T00:31:00.000000000",
                "2012-09-05T00:26:00.000000000",
                "2012-09-05T00:20:00.000000000",
                "2012-09-05T00:15:00.000000000",
                "2012-09-05T00:15:00.000000000",
                "2012-09-05T00:10:00.000000000",
                "2012-09-05T00:04:00.000000000",
                "2012-09-05T00:04:00.000000000",
                "2012-09-04T23:59:00.000000000",
                "2012-09-04T23:54:00.000000000",
                "2012-09-04T23:54:00.000000000",
                "2012-09-04T23:48:00.000000000",
                "2012-09-04T23:43:00.000000000",
                "2012-09-04T23:41:00.000000000",
            ]
        ],
        dtype="datetime64[ns]",
    )

    ex = read_exchange(test_data)

    np.testing.assert_array_equal(ex.bottle_time.values, expected)


def test_empty_string_data():
    """A bug was found in new numpy where an all fill value string param would not convert succesfully"""
    raw = simple_bottle_exchange(params=("HPLC",), units=("",), data=("-999",))
    ex_xr = read_exchange(io.BytesIO(raw))

    assert "hplc_placeholder" in ex_xr.variables
    assert ex_xr.hplc_placeholder.values == [[""]]


def test_alternate_encoding():
    test_str = "Héllø Wörld"
    raw = simple_bottle_exchange(comments=test_str).decode("utf8").encode("latin1")
    with pytest.raises(ExchangeEncodingError):
        ex_xr = read_exchange(io.BytesIO(raw))

    with pytest.raises(ExchangeEncodingError):
        ex_xr = read_exchange(io.BytesIO(raw), encoding="shift-jis")

    ex_xr = read_exchange(io.BytesIO(raw), encoding="latin1")

    assert ex_xr.comments == f"BOTTLE,test\n{test_str}"



import io
import json
import logging
import traceback
from html import escape

from js import Uint8Array, console  # noqa: F401
from pyscript import display

from cchdo.hydro import __version__ as hydro_version
from cchdo.hydro import accessors, read_exchange  # noqa: F401
from cchdo.params import __version__ as params_version


def logger(msg):
    display(msg, target="log", append=True)


class DisplaylHandler(logging.Handler):
    def emit(self, record) -> None:
        logger(self.formatter.format(record))


root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = DisplaylHandler()
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
root.addHandler(handler)


def versions():
    return {
        "hydro_version": hydro_version,
        "params_version": params_version,
    }


def to_netcdf(ex):
    ex.to_netcdf("out.nc")
    with open("out.nc", "rb") as f:
        return f.read()


class Pre:
    def __init__(self, text):
        self.value = text

    def _repr_html_(self):
        return f"<pre>{escape(self.value)}</pre>"


def to_xarray(array_buffer, checks):
    checks = json.loads(checks)
    logger(checks)
    logger("to_xarray called")
    bytes = bytearray(Uint8Array.new(array_buffer))
    logger("got bytes")
    ex_bytes = io.BytesIO(bytes)
    try:
        ex = read_exchange(ex_bytes, checks=checks)
        logger("success! making a netCDF file")
    except ValueError as er:
        display(Pre("".join(traceback.format_exception(er))), target="log", append=True)
        display(er.error_data, target="log", append=True)
        raise  # this is so the promise rejects and the main thread knows what's up
    return to_netcdf(ex), ex.cchdo.gen_fname()


__export__ = ["to_xarray", "versions"]



import json
import logging
import shutil
from collections import Counter
from hashlib import sha256
from html import escape
from multiprocessing import Pool
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import click
import numpy as np
import xarray as xr
from rich.logging import RichHandler
from rich.progress import track

from . import __main_helpers as mh

log = logging.getLogger(__name__)

from . import __version__


def setup_logging(level):
    FORMAT = "%(funcName)s: %(message)s"
    log_handler = RichHandler(level=level)
    logging.basicConfig(
        level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[log_handler]
    )
    logging.captureWarnings(True)


@click.group()
def convert(): ...


def _comment_loader(str_or_path: str) -> str:
    if str_or_path.startswith("@"):
        with open(str_or_path.removeprefix("@")) as f:
            return f.read()
    return str_or_path


PrecisionSouceType = click.Choice(["file", "database"], case_sensitive=False)


@convert.command()
@click.argument("exchange_path")
@click.argument("out_path")
@click.option("--precision_source", default="file", type=PrecisionSouceType)
@click.option("--check-flag/--no-check-flag", default=True)
@click.option(
    "--comments",
    default=None,
    type=str,
    help="either a comment string or file path prefixed with @ (e.g. @README.txt)",
)
def convert_exchange(exchange_path, out_path, check_flag, precision_source, comments):
    setup_logging("DEBUG")
    log.info("Loading read_exchange")
    from .exchange import read_exchange

    checks = {"flags": check_flag}

    ex = read_exchange(exchange_path, checks=checks, precision_source=precision_source)
    log.info("Saving to netCDF")
    if comments is not None:
        comments_contents = _comment_loader(comments)
        ex.attrs["comments"] = comments_contents
    ex.to_netcdf(out_path)
    log.info("Done :)")


@convert.command()
@click.argument("csv_path")
@click.argument("out_path")
@click.option(
    "--ftype",
    default="B",
    type=click.Choice(["B", "C"], case_sensitive=False),
    help="B for Bottle, C for CTD",
)
@click.option("--precision_source", default="file", type=PrecisionSouceType)
@click.option("--check-flag/--no-check-flag", default=True)
@click.option(
    "--comments",
    default=None,
    type=str,
    help="either a comment string or file path prefixed with @ (e.g. @README.txt)",
)
def convert_csv(csv_path, out_path, ftype, check_flag, precision_source, comments):
    setup_logging("DEBUG")
    log.info("Loading read_exchange")
    from .exchange import read_csv

    checks = {"flags": check_flag}

    ex = read_csv(
        csv_path, ftype=ftype, checks=checks, precision_source=precision_source
    )
    log.info("Saving to netCDF")
    if comments is not None:
        comments_contents = _comment_loader(comments)
        ex.attrs["comments"] = comments_contents
    ex.to_netcdf(out_path)
    log.info("Done :)")


@click.group()
def edit(): ...


@edit.command(hidden=True)
@click.argument("expocode")
@click.argument("dtype")
def edit_comments(expocode, dtype):
    from cchdo.auth.session import session as s

    log.info("Loading Cruise Metadata")
    cruises = {
        c["expocode"]: c
        for c in s.get("https://cchdo.ucsd.edu/api/v1/cruise/all").json()
    }
    try:
        cruise = cruises[expocode]
    except KeyError:
        log.error(f"{expocode} not found")
        raise

    cruise_file_ids = set(cruise["files"])

    log.info("Loading Cruise Files")
    files = {f["id"]: f for f in s.get("https://cchdo.ucsd.edu/api/v1/file/all").json()}
    extant_ids = cruise_file_ids & files.keys()
    edit_files = list(
        filter(
            lambda x: x["role"] == "dataset"
            and x["data_format"] == "cf_netcdf"
            and x["data_type"] == dtype,
            (files[id] for id in extant_ids),
        )
    )

    if len(edit_files) == 0:
        log.error("No files to edit")
        exit(1)
    if len(edit_files) > 1:
        log.error("Too many files to edit")
        exit(2)
    file = edit_files[0]
    with NamedTemporaryFile(suffix=".nc") as tf:
        tf.write(s.get(f"https://cchdo.ucsd.edu/{file['file_path']}").content)
        tf.flush()  # when 3.12, switch to delete_on_close=False
        ds = xr.load_dataset(tf.name)

    comments = click.edit(text=ds.attrs.get("comments", ""), extension=".txt")
    if comments is None:
        raise click.Abort

    old_file_id = file["id"]
    old_cruises = file["cruises"]
    new_file = file
    new_file["file_path"] = ""
    new_file["events"] = []
    new_file["submissions"] = []
    del new_file["cruises"]
    del new_file["id"]

    ds.attrs["comments"] = comments
    with TemporaryDirectory() as td:
        temp_nc = td + "/out.nc"
        ds.to_netcdf(temp_nc)
        new_file["file"] = mh.make_netcdf_file_json(temp_nc)
        new_file["file"]["name"] = new_file["file_name"]
        with open(temp_nc, "rb") as f:
            new_file["file_hash"] = sha256(f.read()).hexdigest()
            new_file["file_size"] = f.tell()

    patch = [{"op": "replace", "path": "/role", "value": "merged"}]

    replacement = s.post("https://cchdo.ucsd.edu/api/v1/file", json=new_file)
    print(replacement.json())

    id_ = replacement.json()["message"].split("/")[-1]
    for cruise in old_cruises:
        print(s.post(f"https://cchdo.ucsd.edu/api/v1/cruise/{cruise}/files/{id_}"))

    s.patch(f"https://cchdo.ucsd.edu/api/v1/file/{old_file_id}", json=patch)


@click.group()
def status(): ...


def cchdo_loader(dtype, dformat="exchange"):
    from cchdo.auth.session import session as s

    log.info("Loading Cruise Metadata")
    cruises = s.get("https://cchdo.ucsd.edu/api/v1/cruise/all").json()
    log.info("Loading Cruise File")
    files = s.get("https://cchdo.ucsd.edu/api/v1/file/all").json()

    def file_filter(file):
        return (
            file["data_type"] == dtype
            and file["role"] == "dataset"
            and file["data_format"] == dformat
        )

    return {c["id"]: c for c in cruises}, list(filter(file_filter, files))


def cached_file_loader(file):
    from requests import codes

    from cchdo.auth.session import session as s

    from . import _hydro_appdirs

    cache_dir = Path(_hydro_appdirs.user_cache_dir) / "convert"
    cache_dir.mkdir(parents=True, exist_ok=True)

    file_dest = cache_dir / file["file_hash"]

    if file_dest.exists() and file_dest.stat().st_size == file["file_size"]:
        log.debug(f"Using cached file {file_dest}")
        return file_dest

    f_body = s.get(f"https://cchdo.ucsd.edu{file['file_path']}")
    if f_body.status_code == codes.ok:
        file_dest.write_bytes(f_body.content)
        log.debug(f"Written to {file_dest}")

    return file_dest


def vars_with_value(ds: xr.Dataset) -> list[str]:
    vars_with_data: list[str] = []
    for name, var in ds.variables.items():
        if var.dtype.kind == "f":
            if np.any(np.isfinite(var)).item():
                vars_with_data.append(str(name))
        else:
            vars_with_data.append(str(name))
    return vars_with_data


@status.command()
@click.argument("dtype")
@click.argument("out_dir")
@click.option("--dump-unknown-params", is_flag=True)
@click.option("-v", "--verbose", count=True)
@click.option("--dump-data-counts", is_flag=True)
@click.option("--roundtrip", is_flag=True)
def status_exchange(
    dtype, out_dir, dump_unknown_params, verbose, dump_data_counts, roundtrip
):
    """Generate a bottle conversion status for all ex files of type type in the CCHDO Dataset."""
    from cchdo.hydro import __version__ as hydro_version
    from cchdo.params import __version__ as params_version

    if verbose == 0:
        setup_logging("CRITICAL")
    if verbose == 1:
        setup_logging("INFO")
    if verbose >= 2:
        setup_logging("DEBUG")

    out_path = Path(out_dir)
    cruises, files = cchdo_loader(dtype)
    file_paths = [
        (cached_file_loader(file), file)
        for file in track(files, description="Loading data files")
    ]

    results = []
    all_unknown_params = {}
    variables_with_data = []
    with TemporaryDirectory() as temp_dir:
        with Pool() as pool:
            for result in track(
                pool.imap_unordered(
                    mh.p_file, [(temp_dir, f[0], f[1], roundtrip) for f in file_paths]
                ),
                total=len(file_paths),
                description="Converting ex to netCDF",
            ):
                status, path_or_err, metadata, unknown_params = result
                if status == 200:
                    log.info(f"Processed: {metadata['file_name']}")
                else:
                    log.error(f"Failed: {metadata['file_name']} {path_or_err}")
                results.append(result)

        out_path.mkdir(parents=True, exist_ok=True)
        nc_path = out_path / "nc"
        nc_path.mkdir(exist_ok=True)
        success_len = len(list(filter(lambda x: x[0] == 200, results)))
        success_str = f"Converted {success_len} of {len(results)} ({success_len / len(results) * 100}%)"
        with (out_path / f"index_{dtype}.html").open("w") as f:
            f.write(
                f"""<html>
            <head>
            <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC" crossorigin="anonymous">
            </head>
            <body>
            <div class="container-fluid">
            <table class="table"><thead>
            <h2>Versions</h2>
            cchdo.hydro: {hydro_version}</br>
            cchdo.params: {params_version}</br>
            stats: {success_str}
            <h2>Files</h2>
            <tr>
            <th>Cruise(s)</th>
            <th>File ID</th>
            <th>File Download</th>
            <th>NetCDF File</th>
            </tr></thead><tbody>"""
            )
            for result in sorted(results, key=lambda x: x[2]["id"]):
                status, path_or_err, metadata, unknown_params = result
                try:
                    expos = [cruises[c]["expocode"] for c in metadata["cruises"]]
                    crs = ", ".join(
                        [
                            f"<a href='https://cchdo.ucsd.edu/cruise/{ex}'>{ex}</a>"
                            for ex in expos
                        ]
                    )
                except KeyError:
                    crs = "<span>Not Attached</span>"
                except IndexError:
                    log.critical(metadata["cruises"])
                    raise
                fn = metadata["file_name"]
                file_id = metadata["id"]
                if status == 200:
                    tmp_nc = Path(path_or_err)
                    res_nc = nc_path / tmp_nc.name

                    if dump_data_counts:
                        ds = xr.load_dataset(tmp_nc)
                        variables_with_data.extend(vars_with_value(ds))

                    shutil.copy(tmp_nc, res_nc)
                    f.write(
                        f"""<tr class='table-success'>
                            <td>{crs}</td>
                            <td>{file_id}</td>
                            <td><a href="https://cchdo.ucsd.edu/data/{file_id}/{fn}">{fn}</a></td>
                            <td><a href="nc/{res_nc.name}">{res_nc.name}</a></td>
                            </tr>"""
                    )
                else:
                    if len(unknown_params) > 0:
                        all_unknown_params[expos[0]] = unknown_params
                    error = escape(path_or_err)
                    f.write(
                        f"""<tr class='table-warning'>
                    <td>{crs}</td>
                    <td>{file_id}</td>
                    <td><a href="https://cchdo.ucsd.edu/data/{file_id}/{fn}">{fn}</a></td>"""
                    )

                    if "ExchangeDataFlagPairError" in path_or_err:
                        f.write(
                            f"""<td><pre style="width: fit-content"><code>{error}</code></pre></td>"""
                        )
                    else:
                        f.write(f"""<td>{error}</td>""")

                    f.write("""</tr>""")
            f.write("</tbody></table></div></body></html>")

    log.info(success_str)
    if dump_unknown_params:
        with open(out_path / f"unknown_params_{dtype}.json", "w") as f:
            json.dump(all_unknown_params, f)
    if dump_data_counts:
        with open(out_path / f"params_with_data_{dtype}.json", "w") as f:
            json.dump(Counter(variables_with_data), f)


@status.command()
@click.argument("out_dir")
@click.option("-v", "--verbose", count=True)
@click.option("--only-fail", is_flag=True)
def status_cf_derived(out_dir, verbose, only_fail):
    from cchdo.hydro import __version__ as hydro_version
    from cchdo.params import __version__ as params_version

    if verbose == 0:
        setup_logging("CRITICAL")
    if verbose == 1:
        setup_logging("INFO")
    if verbose >= 2:
        setup_logging("DEBUG")

    out_path = Path(out_dir)
    cruises, bottle_files = cchdo_loader("bottle", "cf_netcdf")
    _, ctd_files = cchdo_loader("ctd", "cf_netcdf")
    all_files = [*bottle_files, *ctd_files]
    file_paths = [
        (cached_file_loader(file), file)
        for file in track(all_files, description="Loading data files")
    ]

    results = []
    with TemporaryDirectory() as temp_dir:
        with Pool() as pool:
            for result in track(
                pool.imap_unordered(
                    mh.p_file_cf, [(temp_dir, f[0], f[1]) for f in file_paths]
                ),
                total=len(file_paths),
                description="Converting CF to Others",
            ):
                metadata, excahnge_ok, coards_ok, woce_ok, sum_ok = result
                log.info(
                    f"Processed: {metadata['file_name']}, {excahnge_ok}, {coards_ok}, {woce_ok}, {sum_ok}"
                )
                results.append(result)

        with (out_path / "index_cf.html").open("w") as f:
            f.write(
                f"""<html>
            <head>
            <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC" crossorigin="anonymous">
            </head>
            <body>
            <div class="container-fluid">
            <table class="table"><thead>
            <h2>Versions</h2>
            cchdo.hydro: {hydro_version}</br>
            cchdo.params: {params_version}</br>
            <h2>Files</h2>
            <tr>
            <th>Cruise(s)</th>
            <th>File ID</th>
            <th>File Download</th>
            <th>Exchange</th>
            <th>COARDS NC</th>
            <th>WOCE</th>
            <th>SUM</th>
            </tr></thead><tbody>"""
            )
            for result in sorted(results, key=lambda x: x[0]["id"]):
                metadata, excahnge_ok, coards_ok, woce_ok, sum_ok = result
                if only_fail and all([excahnge_ok, coards_ok, woce_ok, sum_ok]):
                    continue
                try:
                    expos = [cruises[c]["expocode"] for c in metadata["cruises"]]
                    crs = ", ".join(
                        [
                            f"<a href='https://cchdo.ucsd.edu/cruise/{ex}'>{ex}</a>"
                            for ex in expos
                        ]
                    )
                except KeyError:
                    continue
                except IndexError:
                    log.critical(metadata["cruises"])
                    raise
                fn = metadata["file_name"]
                file_id = metadata["id"]
                f.write(
                    f"""<tr>
                        <td>{crs}</td>
                        <td>{file_id}</td>
                        <td><a href="https://cchdo.ucsd.edu/data/{file_id}/{fn}">{fn}</a></td>
                        <td>{"✅" if excahnge_ok else "❌"}</td>
                        <td>{"✅" if coards_ok else "❌"}</td>
                        <td>{"✅" if woce_ok else "❌"}</td>
                        <td>{"✅" if sum_ok else "❌"}</td>
                        </tr>"""
                )
            f.write("</tbody></table></div></body></html>")


cli = click.version_option(__version__)(
    click.CommandCollection(sources=[convert, status, edit])
)


if __name__ == "__main__":
    cli()



import os
import re
import string
from collections import defaultdict
from datetime import UTC, datetime
from io import BufferedWriter, BytesIO
from typing import Any, Literal, NamedTuple
from zipfile import ZIP_DEFLATED, ZipFile

import numpy as np
import pandas as pd
import xarray as xr

from cchdo.hydro.checks import check_flags as _check_flags
from cchdo.hydro.core import dataarray_factory
from cchdo.hydro.types import FileType
from cchdo.hydro.utils import (
    add_cdom_coordinate,
    all_same,
    extract_numeric_precisions,
    flatten_cdom_coordinate,
)
from cchdo.params import WHPName, WHPNames

FLAG_NAME = "cchdo.hydro._qc"
ERROR_NAME = "cchdo.hydro._error"

PathType = str | bytes | os.PathLike


def write_or_return(
    data: bytes, path_or_fobj: PathType | BufferedWriter | None = None
) -> bytes | None:
    # assume path_or_fobj is an open filelike
    if path_or_fobj is None:
        return data

    if isinstance(path_or_fobj, BufferedWriter):
        try:
            path_or_fobj.write(data)
        except TypeError as error:
            raise TypeError("File must be open for bytes writing") from error
    else:
        with open(path_or_fobj, "wb") as f:
            f.write(data)

    return None


# maybe temp location for FQ merge machinery
class FQPointKey(NamedTuple):
    expocode: str
    station: str
    cast: int
    sample: str


class FQProfileKey(NamedTuple):
    expocode: str
    station: str
    cast: int


class WHPIndxer:
    def __init__(self, obj: xr.Dataset) -> None:
        self.n_prof = pd.MultiIndex.from_arrays(
            [
                obj.expocode.data,
                obj.station.data,
                obj.cast.data,
            ],
            names=["expocode", "station", "cast"],
        )
        self.n_level = []
        for _, prof in obj.groupby("N_PROF", squeeze=False):
            data = prof.sample.squeeze("N_PROF").data
            self.n_level.append(pd.Index(data[data != ""]))

    def __getitem__(self, key: FQProfileKey | FQPointKey):
        prof_idx = self.n_prof.get_loc((key.expocode, key.station, key.cast))
        if isinstance(key, FQPointKey):
            level_idx = self.n_level[prof_idx].get_loc(key.sample)
        else:
            level_idx = slice(None)

        return prof_idx, level_idx


NormalizedFQ = dict[FQProfileKey | FQPointKey, dict[str, str | float]]


def normalize_fq(fq: list[dict[str, str | float]], *, check_dupes=True) -> NormalizedFQ:
    normalized: NormalizedFQ = defaultdict(dict)
    key: FQProfileKey | FQPointKey
    for line in fq:
        line = line.copy()
        expocode = str(line.pop("EXPOCODE"))
        station = str(line.pop("STNNBR"))
        cast = int(line.pop("CASTNO"))
        try:
            sample = str(line.pop("SAMPNO"))
        except KeyError:
            key = FQProfileKey(expocode, station, cast)
        else:
            key = FQPointKey(expocode, station, cast, sample)

        if check_dupes is True:
            shared_keys = normalized[key].keys() & line.keys()
            if len(shared_keys) != 0:
                raise ValueError(f"Duplicate input data found: {key}")

        normalized[key].update(line)

    return normalized


def fq_get_precisions(fq: NormalizedFQ) -> dict[str, int]:
    collect: dict[str, list[str]] = defaultdict(list)
    for value in fq.values():
        for param, data in value.items():
            if isinstance(data, str):
                collect[param].append(data)

    return {
        param: extract_numeric_precisions(data).item()
        for param, data in collect.items()
    }


FTypeOptions = Literal["cf", "exchange", "coards", "woce"]


@xr.register_dataset_accessor("cchdo")
class CCHDOAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        self._obj = xarray_obj

    def jsonld(self) -> dict:
        # we are going to mess with it a little
        obj = self._obj.copy()

        import gsw

        DEPTH = WHPNames["CTDDEPTH [METERS]"]

        base_schema: dict[str, Any] = {
            "@context": {"@vocab": "https://schema.org/"},
            "@id": f"{obj.expocode[0].item()}_{obj.profile_type[0].item()}",  # what would this actually be?
            "@type": "Dataset",
        }
        variableMeasured = []
        NS_TO_S = 1000000000

        # Inject calculated depth if not already present
        if DEPTH.full_nc_name not in obj.variables:
            depth_da = dataarray_factory(
                DEPTH, N_PROF=obj.sizes["N_PROF"], N_LEVELS=obj.sizes["N_LEVELS"]
            )
            depth_da.data = -gsw.z_from_p(obj.pressure, obj.latitude)
            obj[DEPTH.full_nc_name] = depth_da

        minx, maxx, miny, maxy, maxz = None, None, None, None, None
        for var, da in obj.variables.items():
            if "standard_name" not in da.attrs:
                continue
            standard_name = da.attrs["standard_name"]

            if standard_name == "status_flag":
                continue
            variableMeasuredDict = {
                "@type": "PropertyValue",
                "name": var,
                "propertyID": f"http://vocab.nerc.ac.uk/standard_name/{standard_name}",
            }
            if "units" in da.attrs:
                variableMeasuredDict["unitText"] = da.attrs["units"]

            # min/max calculations
            min_v = np.nanmin(da.values).item()
            max_v = np.nanmax(da.values).item()

            # special case time to make it a unix UNIX timestamp... units are ???
            if standard_name == "time":
                min_v = min_v // NS_TO_S
                max_v = max_v // NS_TO_S

            # special case longitude so that "min" and "max" are always westernmost and easternmost
            if standard_name == "longitude" and max_v - min_v > 180:
                max_v = np.nanmax(da.values, where=da.values < 0, initial=-180)
                min_v = np.nanmin(da.values, where=da.values > 0, initial=180)
            if not np.isnan(min_v):
                variableMeasuredDict["minValue"] = min_v
            if not np.isnan(max_v):
                variableMeasuredDict["maxValue"] = max_v
            # if include_description and (sname := CFStandardNames.get(standard_name)) is not None:
            #    variableMeasuredDict["description"] = sname.description

            if standard_name == "longitude":
                minx = min_v
                maxx = max_v
            if standard_name == "latitude":
                miny = min_v
                maxy = max_v
            if standard_name == "depth":
                maxz = max_v

            variableMeasured.append(variableMeasuredDict)
        if (
            minx is not None
            and maxx is not None
            and miny is not None
            and maxy is not None
            and maxz is not None
        ):
            base_schema["spatialCoverage"] = {
                "@type": "Place",
                "geo": {
                    "@type": "GeoShape",
                    "box": f"{miny} {minx} {maxy} {maxx}",
                    "elevation": -int(maxz),
                },
            }

        dtstart = np.min(obj.time).dt.strftime("%Y-%m-%d").item()
        dtend = np.max(obj.time).dt.strftime("%Y-%m-%d").item()
        base_schema["temporalCoverage"] = f"{dtstart}/{dtend}"

        base_schema["variableMeasured"] = variableMeasured
        return base_schema  # TODO pydantic models?

    def to_mat(self, fname):
        """Experimental Matlab .mat data file generator.

        The support for netCDF files in Matlab is really bad.
        Matlab also has no built in support for the standards
        we are trying to follow (CF, ACDD), the most egregious
        lack of support is how to deal with times in netCDF files.
        This was an attempt to make a mat file which takes
        care of some of the things matlab won't do for you.
        It requires scipy to function.

        The file it produces is in no way stable.
        """
        try:
            from scipy.io import savemat as scipy_savemat
        except ImportError as error:
            raise ImportError("scipy is required for mat file saving") from error

        mat_dict = {}
        data = self._obj.to_dict()

        # flatten
        for coord, value in data["coords"].items():
            del value["dims"]
            mat_dict[coord] = value
        for param, value in data["data_vars"].items():
            del value["dims"]
            mat_dict[param] = value

        # cleanups for matlab users
        def to_matdate(dt):
            if dt is None:
                return "NaT"
            return dt.strftime("%d-%b-%Y %H:%M:%S")

        def dt_list_to_str_list(dtl):
            return list(map(to_matdate, dtl))

        for value in mat_dict.values():
            if value.get("attrs", {}).get("standard_name") == "time":
                # the case of list of lists is bottle closure times, which is a sparse array
                if any(isinstance(v, list) for v in value["data"]):
                    value["data"] = list(map(dt_list_to_str_list, value["data"]))
                else:
                    value["data"] = dt_list_to_str_list(value["data"])

            if "status_flag" in value.get("attrs", {}).get("standard_name", ""):
                value["data"] = np.nan_to_num(value["data"], nan=9)

        scipy_savemat(fname, mat_dict)

    def to_coards(self, path=None):
        from .legacy.coards import to_coards

        return write_or_return(to_coards(self._obj), path)

    def to_woce(self, path=None):
        from .legacy.woce import to_woce

        return write_or_return(to_woce(self._obj), path)

    def to_sum(self, path=None):
        """NetCDF to WOCE sumfile maker.

        This is missing some information that is not included anymore (wire out, height above bottom).
        It is especially lacking in including woce parameter IDs
        """
        COMMENTS = "CCHDO SumFile"  # TODO is there a better way?
        SUM_COLUMN_HEADERS_1 = [
            "SHIP/CRS",
            "WOCE",
            "",
            "",
            "CAST",
            "",
            "UTC",
            "EVENT",
            "",
            "POSITION",
            "",
            "UNC",
            "HT ABOVE",
            "WIRE",
            "MAX",
            "NO. OF",
            "",
            "",
        ]
        SUM_COLUMN_HEADERS_2 = [
            "EXPOCODE",
            "SECT",
            "STNNBR",
            "CASTNO",
            "TYPE",
            "DATE",
            "TIME",
            "CODE",
            "LATITUDE",
            "LONGITUDE",
            "NAV",
            "DEPTH",
            "BOTTOM",
            "OUT",
            "PRESS",
            "BOTTLES",
            "PARAMETERS",
            "COMMENTS",
        ]

        SUM_COL_JUSTIFICATION = [  # note that python calls this "align"
            "<",  # expo
            "<",  # woce line
            ">",  # station
            ">",  # cast
            ">",  # type
            "<",  # date
            ">",  # time
            ">",  # EVENT (guess at justification)
            "<",  # position 1
            "<",  # position 2
            "<",  # NAV (guess at justification)
            ">",  # depth
            ">",  # height
            ">",  # wire out
            ">",  # pres
            ">",  # no bottles
            "<",  # params
            "<",  # comments
        ]

        def sum_lat(deg_float):
            deg = abs(int(deg_float))
            if deg_float >= 0:
                hem = "N"
                dec = deg_float % 1
            else:
                hem = "S"
                dec = -(deg_float % -1)

            mins = 60 * dec

            return f"{deg:>2d} {mins:05.2f} {hem}"

        def sum_lon(deg_float):
            deg = abs(int(deg_float))
            if deg_float >= 0:
                hem = "E"
                dec = deg_float % 1
            else:
                hem = "W"
                dec = -(deg_float % -1)

            mins = 60 * dec

            return f"{deg:>3d} {mins:05.2f} {hem}"

        col_widths = [len(s) for s in SUM_COLUMN_HEADERS_1]
        col_widths = [
            max(x, len(s))
            for x, s in zip(col_widths, SUM_COLUMN_HEADERS_2, strict=True)
        ]

        sum_rows = []
        for _, prof in self._obj.groupby("N_PROF", squeeze=False):
            prof = prof.squeeze("N_PROF")
            dt = pd.to_datetime(prof.time.values)

            sect_id = ""
            sect_ids = prof.filter_by_attrs(whp_name="SECT_ID")
            for ids in sect_ids.values():
                sect_id = str(ids.values)
                break

            depth = ""
            depths = prof.filter_by_attrs(whp_name="DEPTH")
            for meters in depths.values():
                depth = f"{meters.values:.0f}"
                if depth == "nan":
                    depth = ""
                break

            no_of_bottles = (
                f"{sum(prof.sample.values != '')}" if prof.profile_type == "B" else ""
            )

            row = [""] * len(col_widths)
            row[0] = str(
                prof.expocode.values
            )  # TODO? Limit to 12 chars as per 3.3.1 of woce manual
            row[1] = sect_id  # TODO? Maybe also limit to 12 chars?
            row[2] = str(prof.station.values)
            row[3] = str(prof.cast.values)
            row[4] = "ROS"
            row[5] = dt.strftime("%m%d%y")
            row[6] = dt.strftime("%H%M")
            row[7] = "BO"
            row[8] = sum_lat(prof.latitude.values)
            row[9] = sum_lon(prof.longitude.values)
            row[10] = "GPS"
            row[11] = depth
            row[12] = ""  # height above "BOTTOM"
            row[13] = ""  # "WIRE" out
            row[14] = f"{max(prof.pressure.values):.0f}"
            row[15] = no_of_bottles
            row[16] = (
                ""  # "PARAMS" we have this info... needs to be calculated on a per profile basis though...
            )
            row[17] = ""  # "COMMENTS"
            sum_rows.append(row)
            col_widths = [max(x, len(s)) for x, s in zip(col_widths, row, strict=True)]

        formats = []
        for width, align in zip(col_widths, SUM_COL_JUSTIFICATION, strict=True):
            formats.append("{: " + align + str(width) + "}")
        format_str = " ".join(formats)

        HEADERS_1 = format_str.format(*SUM_COLUMN_HEADERS_1)
        HEADERS_2 = format_str.format(*SUM_COLUMN_HEADERS_2)
        SEP_LINE = "-" * (sum(col_widths) + len(col_widths))
        SUM_ROWS = [format_str.format(*row) for row in sum_rows]

        sum_file = "\n".join(
            [COMMENTS, HEADERS_1, HEADERS_2, SEP_LINE, *SUM_ROWS]
        ).encode("ascii")
        return write_or_return(sum_file, path)

    @property
    def __geo_interface__(self):
        """The station positions as a MultiPoint geo interface.

        See https://gist.github.com/sgillies/2217756
        """
        ds = self._obj
        coords = np.column_stack((ds.longitude, ds.latitude))

        return {"type": "MultiPoint", "coordinates": coords.tolist()}

    @property
    def track(self):
        """A dict which can be dumped to json which conforms to the expected structure for the CCHDO website."""
        geo = self.__geo_interface__
        if len(geo["coordinates"]) == 1:
            # Website only supports LineString which must contain at least 2 points
            # They can be the same point though
            geo["coordinates"].append(geo["coordinates"][0])

        geo["type"] = "LineString"
        return geo

    @staticmethod
    def _gen_fname(
        expocode: str,
        station: str,
        cast: int,
        profile_type: FileType,
        profile_count: int = 1,
        ftype: FTypeOptions = "cf",
    ) -> str:
        allowed_chars = set(f"._{string.ascii_letters}{string.digits}")

        ctd_one = "ctd.nc"
        ctd_many = "ctd.nc"
        bottle = "bottle.nc"

        if ftype == "exchange":
            ctd_one = "ct1.csv"
            ctd_many = "ct1.zip"
            bottle = "hy1.csv"

        if ftype == "coards":
            # internal zip filenames are done by the legacy writer
            ctd_one = "nc_ctd.zip"
            ctd_many = "nc_ctd.zip"
            bottle = "nc_hyd.zip"

        if ftype == "woce":
            # internal zip filenames are done by the legacy writer
            ctd_one = "ct.txt"
            ctd_many = "ct.zip"
            bottle = "hy.txt"

        if profile_type == FileType.BOTTLE:
            fname = f"{expocode}_{bottle}"
        elif profile_count > 1 or ftype in ("woce", "coards"):
            fname = f"{expocode}_{ctd_many}"
        else:
            fname = f"{expocode}_{station}_{cast:.0f}_{ctd_one}"

        for char in set(fname) - allowed_chars:
            fname = fname.replace(char, "_")

        return fname

    def gen_fname(self, ftype: FTypeOptions = "cf") -> str:
        """Generate a human friendly netCDF (or other output type) filename for this object."""
        expocode = np.atleast_1d(self._obj["expocode"])[0]
        station = np.atleast_1d(self._obj["station"])[0]
        cast = np.atleast_1d(self._obj["cast"])[0]

        profile_type = FileType(np.atleast_1d(self._obj["profile_type"])[0])
        profile_count = len(self._obj.get("N_PROF", []))

        return self._gen_fname(
            expocode, station, cast, profile_type, profile_count, ftype
        )

    def compact_profile(self):
        """Drop the trailing empty data from a profile.

        Because we use the incomplete multidimensional array representation of profiles
        there is often "wasted space" at the end of any profile that is not the longest one.
        This accessor drops that wasted space for xr.Dataset objects containing a single profile
        """
        if self._obj.sizes["N_PROF"] != 1:
            raise NotImplementedError(
                "Cannot compact Dataset with more than one profile"
            )
        return self._obj.isel(N_LEVELS=(self._obj.sample != "")[0])

    date_names = frozenset((WHPNames["DATE"], WHPNames["BTL_DATE"]))
    time_names = frozenset((WHPNames["TIME"], WHPNames["BTL_TIME"]))
    elapsed_time_names = frozenset((WHPNames["CTDETIME [SECONDS]"],))

    @property
    def file_type(self):
        # TODO profile_type is guaranteed to be present
        # TODO profile_type must have C or D as the value
        profile_type = self._obj.profile_type
        if not all_same(profile_type.values):
            raise NotImplementedError(
                "Unable to convert a mix of ctd and bottle (or unknown) dtypes"
            )

        if profile_type[0] == FileType.CTD.value:
            return FileType.CTD
        elif profile_type[0] == FileType.BOTTLE.value:
            return FileType.BOTTLE
        else:
            raise NotImplementedError("Unknown profile type encountered")

    @staticmethod
    def cchdo_c_format_precision(c_format: str) -> int | None:
        if not c_format.endswith("f"):
            return None

        f_format = re.compile(r"\.(\d+)f")
        if (match := f_format.search(c_format)) is not None:
            return int(match.group(1))
        return match

    def _make_params_units_line(
        self,
        params: dict[WHPName, xr.DataArray],
    ):
        plist = []
        ulist = []
        for param, dataarray in sorted(params.items()):
            if self.file_type == FileType.CTD and (
                param.scope != "sample" or param.nc_name == "sample"
            ):
                continue
            plist.append(param.full_whp_name)
            unit = param.whp_unit
            if unit is None:
                unit = ""

            ulist.append(unit)

            if (flag := dataarray.attrs.get(FLAG_NAME)) is not None:
                plist.append(flag.attrs["whp_name"])
                ulist.append("")

            if (error := dataarray.attrs.get(ERROR_NAME)) is not None:
                plist.append(error.attrs["whp_name"])
                ulist.append(unit)

        return ",".join(plist), ",".join(ulist)

    @staticmethod
    def _whpname_from_attrs(attrs) -> list[WHPName]:
        params: list[WHPName] = []
        param = attrs["whp_name"]
        unit = attrs.get("whp_unit")
        if isinstance(param, list):
            params.extend(WHPNames[(combined, unit)] for combined in param)
        else:
            try:
                error = WHPNames[(param, unit)]
                if error.error_col:
                    return []
            except KeyError:
                pass
            params.append(WHPNames[(param, unit)])
        return params

    def _make_ctd_headers(self, params) -> list[str]:
        headers = {}
        for param, da in params.items():
            if param.scope != "profile":
                continue

            date_or_time = None
            if param in self.date_names:
                date_or_time = "date"
                value = da.dt.strftime("%Y%m%d").to_numpy()[0]
            elif param in self.time_names:
                date_or_time = "time"
                value = da.dt.round("min").dt.strftime("%H%M").to_numpy()[0]
            else:
                try:
                    data = da.values[0].item()
                except AttributeError:
                    data = da.values[0]

                numeric_precision_override = self.cchdo_c_format_precision(
                    da.attrs.get("C_format", "")
                )
                value = param.strfex(
                    data,
                    date_or_time=date_or_time,
                    numeric_precision_override=numeric_precision_override,
                )
            headers[param.whp_name] = value.strip()
        return [
            f"NUMBER_HEADERS = {len(headers) + 1}",
            *[f"{key} = {value}" for key, value in headers.items()],
        ]

    def _make_data_block(self, params: dict[WHPName, xr.DataArray]) -> list[str]:
        # TODO N_PROF is guaranteed
        valid_levels = params[WHPNames["SAMPNO"]] != ""
        data_block = []
        for param, da in sorted(params.items()):
            if self.file_type == FileType.CTD and (
                param.scope != "sample" or param.nc_name == "sample"
            ):
                continue
            date_or_time: Literal["date", "time"] | None = None
            # TODO, deal with missing time in BTL_DATE
            if param in self.date_names:
                date_or_time = "date"
                mask = da[valid_levels].isnull()
                values = da[valid_levels].dt.strftime("%Y%m%d").to_numpy()
                values[mask] = "-999"
                values = values.tolist()
            elif param in self.time_names:
                date_or_time = "time"
                mask = da[valid_levels].isnull()
                values = da[valid_levels].dt.round("min").dt.strftime("%H%M").to_numpy()
                values[mask] = "-999"
                values = values.tolist()
            elif param in self.elapsed_time_names and da.dtype.char == "m":
                nat_mask = np.isnat(da.values[valid_levels])
                scale_factor = np.array(1, dtype=da.dtype) / np.array(
                    1, dtype="timedelta64[s]"
                )
                values = da.values[valid_levels].astype("float64") * scale_factor
                values[nat_mask] = np.nan
                values = values.tolist()
            else:
                if da.dtype.char == "m":
                    nat_mask = np.isnat(da)
                    data_t = da.values.astype("timedelta64[s]").astype("float64")
                    data_t[nat_mask] = np.nan
                    data = np.nditer(data_t)
                else:
                    data = np.nditer(da[valid_levels], flags=["refs_ok"])
                numeric_precision_override = self.cchdo_c_format_precision(
                    da.attrs.get("C_format", "")
                )
                values = [
                    param.strfex(
                        v,
                        date_or_time=date_or_time,
                        numeric_precision_override=numeric_precision_override,
                    )
                    for v in data
                ]

            data_block.append(values)

            if (flags := da.attrs.get(FLAG_NAME)) is not None:
                data = np.nditer(flags[valid_levels])
                flag = [param.strfex(v, flag=True) for v in data]
                data_block.append(flag)

            if (errors := da.attrs.get(ERROR_NAME)) is not None:
                data = np.nditer(errors[valid_levels])
                numeric_precision_override = self.cchdo_c_format_precision(
                    da.attrs.get("C_format", "")
                )
                error = [
                    param.strfex(
                        v,
                        date_or_time=date_or_time,
                        numeric_precision_override=numeric_precision_override,
                    )
                    for v in data
                ]
                data_block.append(error)
        return data_block

    def _get_comments(self) -> list[str]:
        if len(comments := self._obj.attrs.get("comments", "")) > 0:
            return [f"#{comment_line}" for comment_line in comments.splitlines()]
        return []

    def to_whp_columns(self, compact=False) -> dict[WHPName, xr.DataArray]:
        # collect all the Exchange variables
        # TODO, all things that appear in an exchange file, must have WHP name
        ds = flatten_cdom_coordinate(self._obj)

        ds = ds.reset_coords(
            [
                "expocode",
                "station",
                "cast",
                "sample",
                "time",
                "latitude",
                "longitude",
                "pressure",
            ]
        )

        ds = ds.stack(ex=("N_PROF", "N_LEVELS"))
        if compact:
            ds = ds.isel(ex=(ds.sample != ""))

        exchange_vars = ds.filter_by_attrs(whp_name=lambda name: name is not None)
        params: dict[WHPName, xr.DataArray] = {}
        for var in exchange_vars.values():
            whp_params = self._whpname_from_attrs(var.attrs)
            for param in whp_params:
                params[param] = var

            ancillary_vars_attr = var.attrs.get("ancillary_variables")
            if ancillary_vars_attr is None:
                continue

            # CF says these need to be space seperated
            ancillary_vars = ancillary_vars_attr.split(" ")
            for ancillary_var in ancillary_vars:
                ancillary = ds[ancillary_var]

                standard_name = ancillary.attrs.get("standard_name")
                if standard_name is None and ancillary.attrs.get("whp_name") is None:
                    # TODO maybe raise...
                    continue

                # currently there are three types of ancillary: flags, errors, and analytical temps (e.g. for pH)
                if standard_name == "temperature_of_analysis_of_sea_water":
                    # this needs to get treated like a param
                    for param in self._whpname_from_attrs(ancillary.attrs):
                        params[param] = ancillary

                elif standard_name == "status_flag":
                    for param in whp_params:
                        ancillary.attrs["whp_name"] = f"{param.full_whp_name}_FLAG_W"
                        params[param].attrs[FLAG_NAME] = ancillary

                # TODO find a way to test this
                try:
                    error_param = WHPNames[
                        (
                            ancillary.attrs.get("whp_name"),
                            ancillary.attrs.get("whp_unit"),
                        )
                    ]
                    if error_param.error_col:
                        ancillary.attrs["whp_name"] = error_param.full_error_name
                        params[param].attrs[ERROR_NAME] = ancillary
                except KeyError:
                    pass

        return params

    def to_exchange(self, path=None):
        """Convert a CCHDO CF netCDF dataset to exchange."""
        # all of the todo comments are for documenting/writing validators
        output_files = {}
        if self.file_type == FileType.CTD:
            for _, ds1 in self._obj.groupby("N_PROF", squeeze=False):
                fname = ds1.cchdo.gen_fname(ftype="exchange")

                output = []
                output.append(f"CTD,{datetime.now(UTC):%Y%m%d}CCHHYDRO")
                output.extend(self._get_comments())

                params = ds1.cchdo.to_whp_columns()
                output.extend(self._make_ctd_headers(params))
                output.extend(self._make_params_units_line(params))

                data_block = self._make_data_block(params)
                output.extend(
                    ",".join(str(cell) for cell in row)
                    for row in zip(*data_block, strict=True)
                )

                output.append("END_DATA\n")
                output_files[fname] = "\n".join(output).encode("utf8")

        if self.file_type == FileType.BOTTLE:
            fname = self._obj.cchdo.gen_fname(ftype="exchange")

            output = []
            output.append(f"BOTTLE,{datetime.now(UTC):%Y%m%d}CCHHYDRO")
            output.extend(self._get_comments())
            params = self._obj.cchdo.to_whp_columns()

            # add the params and units line
            output.extend(self._make_params_units_line(params))

            data_block = self._make_data_block(params)

            for row in zip(*data_block, strict=True):
                output.append(",".join(str(cell) for cell in row))

            output.append("END_DATA\n")
            output_files[fname] = "\n".join(output).encode("utf8")

        if len(output_files) == 1:
            return write_or_return(next(iter(output_files.values())), path)
        output_zip = BytesIO()
        with ZipFile(output_zip, "w", compression=ZIP_DEFLATED) as zipfile:
            for fname, data in output_files.items():
                zipfile.writestr(fname, data)

        output_zip.seek(0)
        return write_or_return(output_zip.read(), path)

    # Until I figure out how to use the pandas machinery (or the explict index project of xarray pays off)
    # I will use a "custom" indexer here to index into the variables
    # This will rely on the N_PROF and N_LEVELS (with extra at some point)
    # * N_PROF will be indexed with (expocode, station, cast)
    # * N_LEVELS will be subindexd with (sample)
    def merge_fq(self, fq: list[dict[str, str | float]], *, check_flags=True):
        # TODOs...
        # * (default True) restrict to open "slots" of non flag 9s
        # * Update history attribute...
        now = datetime.now(UTC)
        new_obj = self._obj.copy(deep=True)
        new_obj = flatten_cdom_coordinate(new_obj)
        idxer = WHPIndxer(new_obj)

        normalized_fq = normalize_fq(fq)
        input_precisions = fq_get_precisions(normalized_fq)
        idxes = {key: idxer[key] for key in normalized_fq}
        # invert keys and indexes?
        inverted: dict[str, dict[str, list]] = defaultdict(lambda: defaultdict(list))
        for key, fq_values in normalized_fq.items():
            for param, value in fq_values.items():
                idx = idxes[key]
                inverted[param]["profs"].append(idx[0])
                inverted[param]["levels"].append(idx[1])
                inverted[param]["values"].append(value)

        for param, values in inverted.items():
            whpname = WHPNames[param]
            if whpname.error_col:
                col_ref = new_obj[whpname.nc_name_error]
            elif whpname.flag_col:
                col_ref = new_obj[whpname.nc_name_flag]
            else:
                col_ref = new_obj[whpname.full_nc_name]

            col_ref.values[values["profs"], values["levels"]] = values["values"]

            col_ref.attrs["date_modified"] = now.isoformat(timespec="seconds")

            if (
                param in input_precisions
                and whpname.dtype == "decimal"
                and not whpname.flag_col
            ):
                new_c_format = f"%.{input_precisions[param]}f"
                new_c_format_source = "input_file"
                if (
                    col_ref.attrs.get("C_format") != new_c_format
                    or col_ref.attrs.get("C_format_source") != new_c_format_source
                ):
                    col_ref.attrs["C_format"] = new_c_format
                    col_ref.attrs["C_format_source"] = new_c_format_source
                    col_ref.attrs["date_metadata_modified"] = now.isoformat(
                        timespec="seconds"
                    )
        new_obj = add_cdom_coordinate(new_obj)
        if check_flags:
            _check_flags(new_obj)
        return new_obj



import logging
from dataclasses import dataclass
from operator import methodcaller

import xarray as xr

# TODO Remove me
from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
# end Remove

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class CheckResult:
    error: str | None = None
    warning: str | None = None

    @property
    def ok(self) -> bool:
        return self.error is None and self.warning is None


class CCHDOnetCDF10:
    __cchdo_version__ = "1.0"

    def check_cf_version(self, ds):
        cf_version = "CF-1.8"
        try:
            conventions = ds.attrs["Conventions"]
        except AttributeError:
            return CheckResult(
                error=f"No 'attrs' attribute on {ds}, is it an xarray.Dataset?"
            )
        except KeyError:
            return CheckResult(error="Global attribute 'Conventions' is required")

        # We need to check that the correct cf conventions version is in the
        # "Conventions" attribute, since there can be more than one whitespace
        # seperated convention, we also need to check the split apart version
        # for exactly the value we expect
        if cf_version in conventions and cf_version in conventions.split():
            return CheckResult()

        return CheckResult(error="{cf_version} not in {conventions}")

    def iter_errors(self, ds: xr.Dataset):
        checks = {
            name: methodcaller(name, ds)
            for name in dir(self)
            if name.startswith("check_")
        }
        log.debug(f"Check methods: {list(checks.keys())}")
        for name, check in checks.items():
            check_result = check(self)
            log.debug(f"{name}: {check_result}")
            if check_result.ok:
                continue
            yield check_result

    def validate(self, ds: xr.Dataset):
        return not any(self.iter_errors(ds))


# Alias the "curren version"
CCHDOnetCDF = CCHDOnetCDF10
